# Orca
#
# Copyright 2023-2025 Igalia, S.L.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=too-many-public-methods
# pylint: disable=too-many-lines

"""Utilities for accessibility inspection."""

from __future__ import annotations

import functools
import inspect
import re
from typing import TYPE_CHECKING, Any

import gi

gi.require_version("Atspi", "2.0")
from gi.repository import Atspi

from . import ax_cache_manager, debug
from .ax_component import AXComponent
from .ax_hypertext import AXHypertext
from .ax_object import AXObject
from .ax_table import AXTable
from .ax_text import AXText
from .ax_utilities_action import AXUtilitiesAction
from .ax_utilities_application import AXUtilitiesApplication
from .ax_utilities_collection import AXUtilitiesCollection
from .ax_utilities_component import AXUtilitiesComponent
from .ax_utilities_document import AXUtilitiesDocument
from .ax_utilities_event import AXUtilitiesEvent
from .ax_utilities_hypertext import AXUtilitiesHypertext
from .ax_utilities_math import AXUtilitiesMath
from .ax_utilities_object import AXUtilitiesObject
from .ax_utilities_relation import AXUtilitiesRelation
from .ax_utilities_role import AXUtilitiesRole
from .ax_utilities_selection import AXUtilitiesSelection
from .ax_utilities_state import AXUtilitiesState
from .ax_utilities_table import AXUtilitiesTable
from .ax_utilities_text import AXUtilitiesText

if TYPE_CHECKING:
    from collections.abc import Callable


class _AXUtilitiesCache:
    """Provides utility-specific access to manager-backed cached values."""

    SET_MEMBERS = "AXUtilities.set-members"
    IS_LAYOUT_ONLY = "AXUtilities.is-layout-only"
    IS_CODE_BLOCK_DESCENDANT = "AXUtilities.is-code-block-descendant"
    IS_COMBO_BOX_DESCENDANT = "AXUtilities.is-combo-box-descendant"
    IS_DOCUMENT_DESCENDANT = "AXUtilities.is-document-descendant"
    IS_EDITABLE_COMBO_BOX_DESCENDANT = "AXUtilities.is-editable-combo-box-descendant"
    IS_EMBEDDED_DESCENDANT = "AXUtilities.is-embedded-descendant"
    IS_ENTRY_DESCENDANT = "AXUtilities.is-entry-descendant"
    IS_GRID_DESCENDANT = "AXUtilities.is-grid-descendant"
    IS_LABEL_OR_CAPTION_DESCENDANT = "AXUtilities.is-label-or-caption-descendant"
    IS_MENU_DESCENDANT = "AXUtilities.is-menu-descendant"
    IS_TOOL_BAR_DESCENDANT = "AXUtilities.is-tool-bar-descendant"
    IS_TOOL_TIP_DESCENDANT = "AXUtilities.is-tool-tip-descendant"
    IS_PRESENTATIONAL_CHILD = "AXUtilities.is-presentational-child"
    IS_TREE_OR_TREE_TABLE_DESCENDANT = "AXUtilities.is-tree-or-tree-table-descendant"
    NEAREST_BLOCK_ANCESTOR = "AXUtilities.nearest-block-ancestor"
    EMBEDDED_DOCUMENT_FRAME = "AXUtilities.embedded-document-frame"

    def __init__(self) -> None:
        self._manager = ax_cache_manager.get_manager()
        for namespace in (
            self.SET_MEMBERS,
            self.IS_LAYOUT_ONLY,
            self.IS_CODE_BLOCK_DESCENDANT,
            self.IS_COMBO_BOX_DESCENDANT,
            self.IS_DOCUMENT_DESCENDANT,
            self.IS_EDITABLE_COMBO_BOX_DESCENDANT,
            self.IS_EMBEDDED_DESCENDANT,
            self.IS_ENTRY_DESCENDANT,
            self.IS_GRID_DESCENDANT,
            self.IS_LABEL_OR_CAPTION_DESCENDANT,
            self.IS_MENU_DESCENDANT,
            self.IS_TOOL_BAR_DESCENDANT,
            self.IS_TOOL_TIP_DESCENDANT,
            self.IS_PRESENTATIONAL_CHILD,
            self.IS_TREE_OR_TREE_TABLE_DESCENDANT,
            self.NEAREST_BLOCK_ANCESTOR,
            self.EMBEDDED_DOCUMENT_FRAME,
        ):
            self._manager.register_cache(
                self,
                namespace,
                lifetime=ax_cache_manager.Lifetime.PROCESS,
                clear_on_demand=ax_cache_manager.ClearPolicy.CLEAR,
            )
        self._members_cache = self._manager.get_cache(self, self.SET_MEMBERS)
        self._layout_only_cache = self._manager.get_cache(self, self.IS_LAYOUT_ONLY)
        self._nearest_block_ancestor_cache = self._manager.get_cache(
            self, self.NEAREST_BLOCK_ANCESTOR
        )
        self._embedded_document_frame_cache = self._manager.get_cache(
            self, self.EMBEDDED_DOCUMENT_FRAME
        )
        self._classification_caches = {
            namespace: self._manager.get_cache(self, namespace)
            for namespace in (
                self.IS_CODE_BLOCK_DESCENDANT,
                self.IS_COMBO_BOX_DESCENDANT,
                self.IS_DOCUMENT_DESCENDANT,
                self.IS_EDITABLE_COMBO_BOX_DESCENDANT,
                self.IS_EMBEDDED_DESCENDANT,
                self.IS_ENTRY_DESCENDANT,
                self.IS_GRID_DESCENDANT,
                self.IS_LABEL_OR_CAPTION_DESCENDANT,
                self.IS_MENU_DESCENDANT,
                self.IS_TOOL_BAR_DESCENDANT,
                self.IS_TOOL_TIP_DESCENDANT,
                self.IS_PRESENTATIONAL_CHILD,
                self.IS_TREE_OR_TREE_TABLE_DESCENDANT,
            )
        }

    def get_members(self, container: Atspi.Accessible | None) -> list[Atspi.Accessible] | None:
        """Returns cached set members for container."""

        if self._members_cache is None:
            return None

        return self._members_cache.get(ax_cache_manager.get_object_key(container), None)

    def set_members(
        self, container: Atspi.Accessible | None, members: list[Atspi.Accessible]
    ) -> None:
        """Stores set members for container."""

        if self._members_cache is not None:
            self._members_cache.put(ax_cache_manager.get_object_key(container), members)

    def get_layout_only(self, obj: Atspi.Accessible) -> tuple[bool, str] | None:
        """Returns the cached layout-only decision for obj."""

        if self._layout_only_cache is None:
            return None

        return self._layout_only_cache.get(ax_cache_manager.get_object_key(obj), None)

    def set_layout_only(self, obj: Atspi.Accessible, result: tuple[bool, str]) -> None:
        """Stores a layout-only decision for obj."""

        if self._layout_only_cache is not None:
            self._layout_only_cache.put(ax_cache_manager.get_object_key(obj), result)

    def get_nearest_block_ancestor(self, obj: Atspi.Accessible) -> Atspi.Accessible | None:
        """Returns the cached nearest block ancestor for obj."""

        if self._nearest_block_ancestor_cache is None:
            return None

        return self._nearest_block_ancestor_cache.get(ax_cache_manager.get_object_key(obj), None)

    def set_nearest_block_ancestor(self, obj: Atspi.Accessible, ancestor: Atspi.Accessible) -> None:
        """Stores the nearest block ancestor for obj."""

        if self._nearest_block_ancestor_cache is not None:
            self._nearest_block_ancestor_cache.put(ax_cache_manager.get_object_key(obj), ancestor)

    def get_embedded_document_frame(self, obj: Atspi.Accessible) -> Any:
        """Returns the cached embedded document frame for obj, or MISSING on a miss."""

        if self._embedded_document_frame_cache is None:
            return ax_cache_manager.MISSING

        key = ax_cache_manager.get_object_key(obj)
        return self._embedded_document_frame_cache.get(key, ax_cache_manager.MISSING)

    def set_embedded_document_frame(
        self, obj: Atspi.Accessible, frame: Atspi.Accessible | None
    ) -> None:
        """Stores the embedded document frame for obj."""

        if self._embedded_document_frame_cache is not None:
            self._embedded_document_frame_cache.put(ax_cache_manager.get_object_key(obj), frame)

    def get_classification(self, namespace: str, obj: Atspi.Accessible) -> bool | None:
        """Returns a cached object classification."""

        cache = self._classification_caches.get(namespace)
        if cache is None:
            return None

        return cache.get(ax_cache_manager.get_object_key(obj), None)

    def set_classification(self, namespace: str, obj: Atspi.Accessible, result: bool) -> None:
        """Stores an object classification."""

        cache = self._classification_caches.get(namespace)
        if cache is not None:
            cache.put(ax_cache_manager.get_object_key(obj), result)


class AXUtilities:
    """Utilities for accessibility inspection."""

    COMPARE_COLLECTION_PERFORMANCE = False

    _CACHE = _AXUtilitiesCache()

    @staticmethod
    def clear_all_cache_now(obj: Atspi.Accessible | None = None, reason: str = "") -> None:
        """Clears all cached information immediately."""

        ax_cache_manager.get_manager().clear_cache_now(reason)
        if AXUtilitiesRole.is_table_related(obj):
            ax_cache_manager.get_manager().invalidate_group(
                AXTable.CACHE_INVALIDATION_GROUP, reason
            )

    @staticmethod
    def can_be_active_window(window: Atspi.Accessible, clear_cache: bool = True) -> bool:
        """Returns True if window can be the active window based on its state."""

        if window is None:
            return False

        if clear_cache:
            AXObject.clear_cache(window, False, "Checking if window can be the active window")
        app = AXUtilitiesApplication.get_application(window)
        tokens = ["AXUtilities:", window, "from", app]

        window_states = AXObject.get_state_set(window)
        can_be_active = True
        if not AXUtilitiesState.is_active(window, window_states):
            tokens.append("lacks active state")
            can_be_active = False
        elif not AXUtilitiesState.is_showing(window, window_states):
            tokens.append("lacks showing state")
            can_be_active = False
        elif AXUtilitiesState.is_iconified(window, window_states):
            tokens.append("is iconified")
            can_be_active = False
        elif AXUtilitiesApplication.is_mutter_x11_frames(app):
            tokens.append("is from app that cannot have the real active window")
            can_be_active = False
        elif app and not AXUtilitiesApplication.is_application_in_desktop(app):
            tokens.append("is from app unknown to AT-SPI2")
            # Firefox alerts and dialogs suffer from this bug too, but if we ignore these windows
            # we'll fail to fully present things like the file chooser dialog and the replace-file
            # alert. https://bugzilla.mozilla.org/show_bug.cgi?id=1882794
            if not AXUtilitiesRole.is_dialog_or_alert(window):
                can_be_active = False

        if can_be_active:
            tokens.append("can be active window")
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return can_be_active

    @staticmethod
    def find_active_window() -> Atspi.Accessible | None:
        """Tries to locate the active window; may or may not succeed."""

        candidates = []
        apps = AXUtilitiesApplication.get_all_applications(must_have_window=True)
        for app in apps:
            candidates.extend(list(AXObject.iter_children(app, AXUtilities.can_be_active_window)))

        if not candidates:
            tokens = ["AXUtilities: Unable to find active window from", apps]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return None

        if len(candidates) == 1:
            tokens = ["AXUtilities: Active window is", candidates[0]]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return candidates[0]

        tokens = ["AXUtilities: These windows all claim to be active:", candidates]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        # Some electron apps running in the background claim to be active even when they
        # are not. These are the ones we know about. We can add others as we go.
        suspect_apps = ["slack", "discord", "outline-client", "whatsapp-desktop-linux"]
        filtered = []
        for frame in candidates:
            if AXObject.get_name(AXUtilitiesApplication.get_application(frame)) in suspect_apps:
                tokens = ["AXUtilities: Suspecting", frame, "is a non-active Electron app"]
                debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            else:
                filtered.append(frame)

        if len(filtered) == 1:
            tokens = ["AXUtilities: Active window is believed to be", filtered[0]]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return filtered[0]

        guess: Atspi.Accessible | None = None
        if filtered:
            tokens = ["AXUtilities: Still have multiple active windows:", filtered]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            guess = filtered[0]

        if guess is not None:
            tokens = ["AXUtilities: Returning", guess, "as active window"]
        else:
            tokens = ["AXUtilities: No active window found"]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return guess

    @staticmethod
    def is_unfocused_alert_or_dialog(obj: Atspi.Accessible) -> bool:
        """Returns True if obj is an unfocused alert or dialog with presentable items."""

        if not AXUtilitiesRole.is_dialog_or_alert(obj):
            return False
        if not AXObject.get_child_count(obj):
            return False
        if not AXUtilitiesState.is_showing(obj):
            return False
        return not AXUtilities.can_be_active_window(obj)

    @staticmethod
    def get_unfocused_alerts_and_dialogs(obj: Atspi.Accessible) -> list[Atspi.Accessible]:
        """Returns a list of all the unfocused alerts and dialogs in the app and window of obj."""

        app = AXUtilitiesApplication.get_application(obj)
        result = list(AXObject.iter_children(app, AXUtilities.is_unfocused_alert_or_dialog))

        frame = AXUtilitiesObject.find_ancestor(
            obj,
            lambda x: AXUtilitiesRole.is_application(AXObject.get_parent(x)),
        )
        result.extend(list(AXObject.iter_children(frame, AXUtilities.is_unfocused_alert_or_dialog)))

        tokens = ["AXUtilities: Unfocused alerts and dialogs for", obj, ":", result]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return result

    @staticmethod
    def get_all_widgets(
        obj: Atspi.Accessible,
        must_be_showing_and_visible: bool = True,
        exclude_push_button: bool = False,
    ) -> list[Atspi.Accessible]:
        """Returns all the descendants of obj with a widget role"""

        roles = AXUtilitiesRole.get_widget_roles()
        if exclude_push_button and Atspi.Role.BUTTON in roles:
            roles.remove(Atspi.Role.BUTTON)

        result = None
        if AXObject.supports_collection(obj):
            if not must_be_showing_and_visible:
                result = AXUtilitiesCollection.find_all_with_role(obj, roles)
            else:
                states = [Atspi.StateType.SHOWING, Atspi.StateType.VISIBLE]
                result = AXUtilitiesCollection.find_all_with_role_and_all_states(obj, roles, states)

            if not AXUtilities.COMPARE_COLLECTION_PERFORMANCE:
                return result

        def is_match(acc):
            if AXObject.get_role(acc) not in roles:
                return False
            if must_be_showing_and_visible:
                return AXUtilitiesState.is_showing_and_visible(acc)
            return True

        return AXUtilitiesObject.find_all_descendants(obj, is_match)

    @staticmethod
    def get_default_button(obj: Atspi.Accessible) -> Atspi.Accessible | None:
        """Returns the default button descendant of obj"""

        result = None
        if AXObject.supports_collection(obj):
            result = AXUtilitiesCollection.find_default_button(obj)
            if not AXUtilities.COMPARE_COLLECTION_PERFORMANCE:
                return result

        return AXUtilitiesObject.find_descendant(obj, AXUtilitiesRole.is_default_button)

    @staticmethod
    def get_details_content(obj: Atspi.Accessible) -> list[str]:
        """Returns a list of non-empty text strings from the details of obj."""

        details = AXUtilitiesRelation.get_details(obj)
        if (
            not details
            and AXUtilitiesRole.is_toggle_button(obj)
            and AXUtilitiesState.is_expanded(obj)
        ):
            details = list(AXObject.iter_children(obj))

        result: list[str] = []
        for detail in details:
            texts = [
                AXText.get_all_text(d).strip() for d in AXUtilities.get_text_descendants(detail)
            ]
            result.extend(t for t in texts if t)
        return result

    @staticmethod
    def get_description_list_terms(obj: Atspi.Accessible) -> list[Atspi.Accessible]:
        """Returns all description list terms in obj, excluding nested description lists."""

        if not AXUtilitiesRole.is_description_list(obj):
            return []

        if AXObject.supports_collection(obj):
            matches = AXUtilitiesCollection.find_all_description_terms(obj)
            return [
                m
                for m in matches
                if AXUtilitiesObject.find_ancestor(m, AXUtilitiesRole.is_description_list) == obj
            ]

        return AXUtilitiesObject.find_all_descendants(
            obj, AXUtilitiesRole.is_description_term, AXUtilitiesRole.is_description_list
        )

    @staticmethod
    def get_text_descendants(obj: Atspi.Accessible) -> list[Atspi.Accessible]:
        """Returns all descendants of obj that implement the Text interface."""

        if AXObject.supports_collection(obj):
            return AXUtilitiesCollection.find_all_with_interfaces(obj, ["Text"])

        return AXUtilitiesObject.find_all_descendants(obj, AXObject.supports_text)

    @staticmethod
    def get_focused_object(obj: Atspi.Accessible) -> Atspi.Accessible | None:
        """Returns the focused descendant of obj"""

        result = None
        if AXObject.supports_collection(obj):
            result = AXUtilitiesCollection.find_focused_object(obj)
            if not AXUtilities.COMPARE_COLLECTION_PERFORMANCE:
                return result

        return AXUtilitiesObject.find_descendant(obj, AXUtilitiesState.is_focused)

    @staticmethod
    def get_info_bar(obj: Atspi.Accessible) -> Atspi.Accessible | None:
        """Returns the info bar descendant of obj"""

        result = None
        if AXObject.supports_collection(obj):
            result = AXUtilitiesCollection.find_info_bar(obj)
            if not AXUtilities.COMPARE_COLLECTION_PERFORMANCE:
                return result

        return AXUtilitiesObject.find_descendant(obj, AXUtilitiesRole.is_info_bar)

    @staticmethod
    def get_status_bar(obj: Atspi.Accessible) -> Atspi.Accessible | None:
        """Returns the status bar descendant of obj"""

        result = None
        if AXObject.supports_collection(obj):
            result = AXUtilitiesCollection.find_status_bar(obj)
            if not AXUtilities.COMPARE_COLLECTION_PERFORMANCE:
                return result

        return AXUtilitiesObject.find_descendant(obj, AXUtilitiesRole.is_status_bar)

    @staticmethod
    def get_label(obj: Atspi.Accessible) -> Atspi.Accessible | None:
        """Returns the label descendant of obj"""

        result = None
        if AXObject.supports_collection(obj):
            result = AXUtilitiesCollection.find_first_with_role(obj, [Atspi.Role.LABEL])
            if not AXUtilities.COMPARE_COLLECTION_PERFORMANCE:
                return result

        return AXUtilitiesObject.find_descendant(obj, AXUtilitiesRole.is_label)

    @staticmethod
    def get_list_item(obj: Atspi.Accessible) -> Atspi.Accessible | None:
        """Returns the list item descendant of obj"""

        result = None
        if AXObject.supports_collection(obj):
            result = AXUtilitiesCollection.find_first_with_role(obj, [Atspi.Role.LIST_ITEM])
            if not AXUtilities.COMPARE_COLLECTION_PERFORMANCE:
                return result

        return AXUtilitiesObject.find_descendant(obj, AXUtilitiesRole.is_list_item)

    @staticmethod
    def get_menu(obj: Atspi.Accessible) -> Atspi.Accessible | None:
        """Returns the menu descendant of obj"""

        result = None
        if AXObject.supports_collection(obj):
            result = AXUtilitiesCollection.find_first_with_role(obj, [Atspi.Role.MENU])
            if not AXUtilities.COMPARE_COLLECTION_PERFORMANCE:
                return result

        return AXUtilitiesObject.find_descendant(obj, AXUtilitiesRole.is_menu)

    @staticmethod
    def get_menu_or_list_box(obj: Atspi.Accessible) -> Atspi.Accessible | None:
        """Returns the menu or list box descendant of obj"""

        result = None
        if AXObject.supports_collection(obj):
            result = AXUtilitiesCollection.find_first_with_role(
                obj, [Atspi.Role.MENU, Atspi.Role.LIST_BOX]
            )
            if not AXUtilities.COMPARE_COLLECTION_PERFORMANCE:
                return result

        return AXUtilitiesObject.find_descendant(
            obj, lambda x: AXUtilitiesRole.is_menu(x) or AXUtilitiesRole.is_list_box(x)
        )

    @staticmethod
    def get_progress_bar(obj: Atspi.Accessible) -> Atspi.Accessible | None:
        """Returns the progress bar descendant of obj"""

        result = None
        if AXObject.supports_collection(obj):
            result = AXUtilitiesCollection.find_first_with_role(obj, [Atspi.Role.PROGRESS_BAR])
            if not AXUtilities.COMPARE_COLLECTION_PERFORMANCE:
                return result

        return AXUtilitiesObject.find_descendant(obj, AXUtilitiesRole.is_progress_bar)

    @staticmethod
    def get_slider(obj: Atspi.Accessible) -> Atspi.Accessible | None:
        """Returns the slider descendant of obj"""

        result = None
        if AXObject.supports_collection(obj):
            result = AXUtilitiesCollection.find_first_with_role(obj, [Atspi.Role.SLIDER])
            if not AXUtilities.COMPARE_COLLECTION_PERFORMANCE:
                return result

        return AXUtilitiesObject.find_descendant(obj, AXUtilitiesRole.is_slider)

    @staticmethod
    def get_table_cell(obj: Atspi.Accessible) -> Atspi.Accessible | None:
        """Returns the table cell descendant of obj"""

        result = None
        if AXObject.supports_collection(obj):
            result = AXUtilitiesCollection.find_first_with_role(obj, [Atspi.Role.TABLE_CELL])
            if not AXUtilities.COMPARE_COLLECTION_PERFORMANCE:
                return result

        return AXUtilitiesObject.find_descendant(obj, AXUtilitiesRole.is_table_cell)

    @staticmethod
    def get_table_cell_or_header(obj: Atspi.Accessible) -> Atspi.Accessible | None:
        """Returns the table cell or header descendant of obj"""

        result = None
        if AXObject.supports_collection(obj):
            result = AXUtilitiesCollection.find_first_with_role(
                obj, AXUtilitiesRole.get_table_cell_roles()
            )
            if not AXUtilities.COMPARE_COLLECTION_PERFORMANCE:
                return result

        return AXUtilitiesObject.find_descendant(obj, AXUtilitiesRole.is_table_cell_or_header)

    @staticmethod
    def get_descendant_supporting_text(obj: Atspi.Accessible) -> Atspi.Accessible | None:
        """Returns the first descendant of obj implementing the text interface"""

        result = None
        if AXObject.supports_collection(obj):
            result = AXUtilitiesCollection.find_first_with_interfaces(obj, ["Text"])
            if not AXUtilities.COMPARE_COLLECTION_PERFORMANCE:
                return result

        return AXUtilitiesObject.find_descendant(obj, AXObject.supports_text)

    @staticmethod
    def get_text_input(obj: Atspi.Accessible) -> Atspi.Accessible | None:
        """Returns the text input descendant of obj"""

        result = None
        if AXObject.supports_collection(obj):
            result = AXUtilitiesCollection.find_first_with_role(
                obj, [Atspi.Role.ENTRY, Atspi.Role.PASSWORD_TEXT, Atspi.Role.SPIN_BUTTON]
            )
            if result is None:
                result = AXUtilitiesCollection.find_first_with_role_and_all_states(
                    obj, [Atspi.Role.TEXT], [Atspi.StateType.EDITABLE, Atspi.StateType.SINGLE_LINE]
                )
            if not AXUtilities.COMPARE_COLLECTION_PERFORMANCE:
                return result

        return AXUtilitiesObject.find_descendant(obj, AXUtilitiesRole.is_text_input)

    @staticmethod
    def _is_layout_only_layered_pane(obj: Atspi.Accessible) -> tuple[bool, str]:
        """Returns True with reason if this layered pane is layout-only."""

        if AXUtilitiesObject.find_ancestor(obj, AXUtilitiesRole.is_desktop_frame) is not None:
            return True, "is inside desktop frame"
        return False, ""

    @staticmethod
    def _is_layout_only_menu_or_list(obj: Atspi.Accessible) -> tuple[bool, str]:
        """Returns True with reason if this menu or list in a combo box is layout-only."""

        if AXUtilities.is_combo_box_descendant(obj):
            return True, "is inside combo box"
        return False, ""

    @staticmethod
    def _is_layout_only_group(obj: Atspi.Accessible) -> tuple[bool, str]:
        """Returns True with reason if this group is layout-only."""

        if not AXUtilities.has_explicit_name(obj):
            return True, "lacks explicit name"
        return False, ""

    @staticmethod
    def _is_layout_only_panel(obj: Atspi.Accessible) -> tuple[bool, str]:
        """Returns True with reason if this panel or grouping is layout-only."""

        name = AXObject.get_name(obj)
        description = AXObject.get_description(obj)
        labelled_by = AXUtilitiesRelation.get_is_labelled_by(obj)
        described_by = AXUtilitiesRelation.get_is_described_by(obj)
        if not (name or description or labelled_by or described_by):
            return True, "lacks name, description, and relations"
        if name == AXObject.get_name(AXUtilitiesApplication.get_application(obj)):
            return True, "has same name as app"
        if AXObject.get_child_count(obj) == 1:
            child = AXObject.get_child(obj, 0)
            if name == AXObject.get_name(child):
                return True, "has same name as its only child"
            if not AXUtilitiesRole.is_label(child) and child in labelled_by:
                return True, "is labelled by non-label only child"
        set_roles = AXUtilitiesRole.get_set_container_roles()
        ancestor = AXUtilitiesObject.find_ancestor(obj, lambda x: AXObject.get_role(x) in set_roles)
        if ancestor and AXObject.get_name(ancestor) == name:
            return True, "is in set container with same name"
        return False, ""

    @staticmethod
    def _is_layout_only_section(obj: Atspi.Accessible) -> tuple[bool, str]:
        """Returns True with reason if this section or document is layout-only."""

        if AXUtilitiesState.is_focusable(obj):
            return False, "is focusable"
        if AXUtilitiesAction.has_action(obj, "click"):
            return False, "has click action"
        if AXUtilitiesRole.is_code_block(obj):
            return False, "is code block"
        return True, "is not interactive"

    @staticmethod
    def _is_layout_only_toolbar(obj: Atspi.Accessible) -> tuple[bool, str]:
        """Returns True with reason if this toolbar is layout-only."""

        if AXUtilitiesRole.is_page_tab_list(AXObject.get_child(obj, 0)):
            return True, "is parent of page tab list"
        return False, ""

    @staticmethod
    def _is_layout_only_table(obj: Atspi.Accessible) -> tuple[bool, str]:
        """Returns True with reason if this table is layout-only."""

        if AXUtilitiesTable.is_layout_table(obj):
            return True, "is layout table"
        return False, ""

    @staticmethod
    def _is_layout_only_table_row(obj: Atspi.Accessible) -> tuple[bool, str]:
        """Returns True with reason if this table row is layout-only."""

        state_set = AXObject.get_state_set(obj)
        if AXUtilitiesState.is_focusable(obj, state_set):
            return False, "is focusable"
        if AXUtilitiesState.is_selectable(obj, state_set):
            return False, "is selectable"
        if AXUtilitiesState.is_expandable(obj, state_set):
            return False, "is expandable"
        if AXUtilities.has_explicit_name(obj):
            return False, "has explicit name"
        return True, "is not focusable, selectable, or expandable and lacks explicit name"

    @staticmethod
    def _is_layout_only_table_cell(obj: Atspi.Accessible) -> tuple[bool, str]:
        """Returns True with reason if this table cell is layout-only."""

        if AXUtilitiesRole.is_table_cell(AXObject.get_child(obj, 0)):
            return True, "child of this cell is table cell"
        table = AXUtilitiesTable.get_table(obj)
        if AXUtilitiesRole.is_table(table) and AXUtilitiesTable.is_layout_table(table):
            return True, "is in layout table"
        return False, ""

    @staticmethod
    def _is_layout_only(obj: Atspi.Accessible) -> tuple[bool, str]:
        """Returns True and a string reason if obj is believed to serve only for layout."""

        role = AXObject.get_role(obj)
        if role in AXUtilitiesRole.get_layout_only_roles():
            return True, "has layout-only role"

        result, reason = False, ""
        if AXUtilitiesRole.is_layered_pane(obj, role):
            result, reason = AXUtilities._is_layout_only_layered_pane(obj)
        elif AXUtilitiesRole.is_menu(obj, role) or AXUtilitiesRole.is_list(obj, role):
            result, reason = AXUtilities._is_layout_only_menu_or_list(obj)
        elif AXUtilitiesRole.is_group(obj, role):
            result, reason = AXUtilities._is_layout_only_group(obj)
        elif AXUtilitiesRole.is_panel(obj, role) or AXUtilitiesRole.is_grouping(obj, role):
            result, reason = AXUtilities._is_layout_only_panel(obj)
        elif AXUtilities.is_embedded_document_frame(obj):
            result, reason = False, "is an embedded document frame"
        elif AXUtilitiesRole.is_section(obj, role) or AXUtilitiesRole.is_document(obj, role):
            result, reason = AXUtilities._is_layout_only_section(obj)
        elif AXUtilitiesRole.is_tool_bar(obj):
            result, reason = AXUtilities._is_layout_only_toolbar(obj)
        elif AXUtilitiesRole.is_table(obj, role):
            result, reason = AXUtilities._is_layout_only_table(obj)
        elif AXUtilitiesRole.is_table_row(obj):
            result, reason = AXUtilities._is_layout_only_table_row(obj)
        elif AXUtilitiesRole.is_table_cell(obj, role):
            result, reason = AXUtilities._is_layout_only_table_cell(obj)
        return result, reason

    @staticmethod
    def is_layout_only(obj: Atspi.Accessible) -> bool:
        """Returns True if obj is believed to serve only for layout."""

        cached = AXUtilities._CACHE.get_layout_only(obj)
        if cached is None:
            result, reason = AXUtilities._is_layout_only(obj)
            AXUtilities._CACHE.set_layout_only(obj, (result, reason))
        else:
            result, reason = cached

        if reason:
            tokens = ["AXUtilities:", obj, f"believed to be layout only: {result}, {reason}"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        return result

    @staticmethod
    def _is_descendant(
        namespace: str,
        obj: Atspi.Accessible,
        pred: Callable[[Atspi.Accessible], bool],
        inclusive: bool = False,
    ) -> bool:
        """Returns True if obj (or an ancestor) matches pred."""

        if inclusive and pred(obj):
            return True

        rv = AXUtilities._CACHE.get_classification(namespace, obj)
        if rv is not None:
            return rv

        parent = AXObject.get_parent(obj)
        if parent is None:
            AXUtilities._CACHE.set_classification(namespace, obj, False)
            return False

        if pred(parent):
            AXUtilities._CACHE.set_classification(namespace, obj, True)
            return True

        parent_rv = AXUtilities._CACHE.get_classification(namespace, parent)
        if parent_rv is None:
            parent_rv = AXUtilitiesObject.find_ancestor(parent, pred) is not None
            AXUtilities._CACHE.set_classification(namespace, parent, parent_rv)

        AXUtilities._CACHE.set_classification(namespace, obj, parent_rv)
        return parent_rv

    @staticmethod
    def is_code_block_descendant(
        obj: Atspi.Accessible,
        inclusive: bool = False,
    ) -> bool:
        """Returns True if obj has a code block ancestor."""

        return AXUtilities._is_descendant(
            AXUtilities._CACHE.IS_CODE_BLOCK_DESCENDANT,
            obj,
            AXUtilitiesRole.is_code_block,
            inclusive,
        )

    @staticmethod
    def is_custom_image(obj: Atspi.Accessible) -> bool:
        """Returns True if obj is a custom web element acting as an image."""

        if not AXUtilitiesRole.is_web_element_custom(obj):
            return False

        if not (
            AXUtilitiesRole.is_section(obj)
            and AXUtilities.has_explicit_name(obj)
            and AXObject.supports_text(obj)
            and not re.search(r"[^\s\ufffc]", AXText.get_all_text(obj))
        ):
            return False

        return all(
            AXUtilitiesRole.is_image_or_canvas(child) or AXUtilitiesRole.is_svg(child)
            for child in AXObject.iter_children(obj)
        )

    @staticmethod
    def is_text_block(
        obj: Atspi.Accessible,
        role: Atspi.Role | None = None,
        exclude_editable: bool = False,
        exclude_focusable: bool = False,
    ) -> bool:
        """Returns True if obj is a non-interactive text block."""

        if not AXObject.supports_text(obj):
            return False
        if exclude_editable and AXUtilitiesState.is_editable(obj):
            return False
        if exclude_focusable and AXUtilitiesState.is_focusable(obj):
            return False
        if AXUtilities.is_custom_image(obj):
            return False
        if role is None:
            role = AXObject.get_role(obj)
        if role == Atspi.Role.TABLE_CELL:
            return not AXUtilitiesRole.is_grid_cell(obj)
        return role in AXUtilitiesRole.get_text_block_roles()  # pylint: disable=unsupported-membership-test

    @staticmethod
    def is_combo_box_descendant(
        obj: Atspi.Accessible,
        inclusive: bool = False,
    ) -> bool:
        """Returns True if obj has a combo box ancestor."""

        return AXUtilities._is_descendant(
            AXUtilities._CACHE.IS_COMBO_BOX_DESCENDANT,
            obj,
            AXUtilitiesRole.is_combo_box,
            inclusive,
        )

    @staticmethod
    def is_document_descendant(
        obj: Atspi.Accessible,
        inclusive: bool = False,
    ) -> bool:
        """Returns True if obj has a document ancestor."""

        return AXUtilities._is_descendant(
            AXUtilities._CACHE.IS_DOCUMENT_DESCENDANT,
            obj,
            AXUtilitiesRole.is_document,
            inclusive,
        )

    @staticmethod
    def is_editable_combo_box_descendant(
        obj: Atspi.Accessible,
        inclusive: bool = False,
    ) -> bool:
        """Returns True if obj has an editable combo box ancestor."""

        return AXUtilities._is_descendant(
            AXUtilities._CACHE.IS_EDITABLE_COMBO_BOX_DESCENDANT,
            obj,
            AXUtilitiesRole.is_editable_combo_box,
            inclusive,
        )

    @staticmethod
    def is_embedded_descendant(
        obj: Atspi.Accessible,
        inclusive: bool = False,
    ) -> bool:
        """Returns True if obj has an embedded ancestor."""

        return AXUtilities._is_descendant(
            AXUtilities._CACHE.IS_EMBEDDED_DESCENDANT,
            obj,
            AXUtilitiesRole.is_embedded,
            inclusive,
        )

    @staticmethod
    def is_embedded_document_frame(obj: Atspi.Accessible) -> bool:
        """Returns True if obj is a document frame nested inside an application."""

        return AXUtilitiesRole.is_document_frame(obj) and AXUtilities.is_embedded_descendant(obj)

    @staticmethod
    def is_entry_descendant(
        obj: Atspi.Accessible,
        inclusive: bool = False,
    ) -> bool:
        """Returns True if obj has an entry ancestor."""

        return AXUtilities._is_descendant(
            AXUtilities._CACHE.IS_ENTRY_DESCENDANT,
            obj,
            AXUtilitiesRole.is_entry,
            inclusive,
        )

    @staticmethod
    def get_nearest_block_ancestor(
        obj: Atspi.Accessible, original: Atspi.Accessible | None = None
    ) -> Atspi.Accessible | None:
        """Returns obj or its nearest ancestor that is not an inline element."""

        if original is None:
            original = obj

        # The wrapper transparency below holds only while ascending from the inline start, not
        # for the start itself; the start is kept out of the cache too, so its answer and the
        # ancestor answer for the same object cannot overwrite one another.
        above_start = obj is not original
        if above_start:
            rv = AXUtilities._CACHE.get_nearest_block_ancestor(obj)
            if rv is not None:
                return rv

        inline = AXUtilitiesRole.is_inline_element(obj)
        parent = AXObject.get_parent(obj)
        if not inline and above_start and AXUtilitiesRole.is_section(obj):
            # A generic section whose text is only embedded objects (e.g. like-button-view-
            # model) is a transparent wrapper; the section gate keeps a structural [OBJ]-only
            # element (a table row of cells) from being treated as a wrapper.
            text = AXText.get_all_text(obj)
            inline = "\ufffc" in text and not re.search(r"[^\s\ufffc]", text)

        if inline:
            rv = (
                AXUtilities.get_nearest_block_ancestor(parent, original)
                if parent is not None
                else obj
            )
        else:
            rv = obj

        if above_start and rv is not None:
            AXUtilities._CACHE.set_nearest_block_ancestor(obj, rv)
        return rv

    @staticmethod
    def get_embedded_document_frame_for_object(obj: Atspi.Accessible) -> Atspi.Accessible | None:
        """Returns obj's nearest document-frame ancestor that is inside an application, if any."""

        cached = AXUtilities._CACHE.get_embedded_document_frame(obj)
        if cached is not ax_cache_manager.MISSING:
            return cached

        rv = AXUtilitiesObject.find_ancestor_inclusive(obj, AXUtilities.is_embedded_document_frame)
        AXUtilities._CACHE.set_embedded_document_frame(obj, rv)
        return rv

    @staticmethod
    def is_grid_descendant(
        obj: Atspi.Accessible,
        inclusive: bool = False,
    ) -> bool:
        """Returns True if obj has a grid ancestor."""

        return AXUtilities._is_descendant(
            AXUtilities._CACHE.IS_GRID_DESCENDANT,
            obj,
            AXUtilitiesRole.is_grid,
            inclusive,
        )

    @staticmethod
    def is_label_or_caption_descendant(
        obj: Atspi.Accessible,
        inclusive: bool = False,
    ) -> bool:
        """Returns True if obj has a label or caption ancestor."""

        return AXUtilities._is_descendant(
            AXUtilities._CACHE.IS_LABEL_OR_CAPTION_DESCENDANT,
            obj,
            AXUtilitiesRole.is_label_or_caption,
            inclusive,
        )

    @staticmethod
    def is_menu_descendant(
        obj: Atspi.Accessible,
        inclusive: bool = False,
    ) -> bool:
        """Returns True if obj has a menu ancestor."""

        return AXUtilities._is_descendant(
            AXUtilities._CACHE.IS_MENU_DESCENDANT,
            obj,
            AXUtilitiesRole.is_menu,
            inclusive,
        )

    @staticmethod
    def is_presentational_child(obj: Atspi.Accessible) -> bool:
        """Returns True if obj has an ancestor with presentational children."""

        return AXUtilities._is_descendant(
            AXUtilities._CACHE.IS_PRESENTATIONAL_CHILD,
            obj,
            AXUtilitiesRole.children_are_presentational,
        )

    @staticmethod
    def is_tool_bar_descendant(
        obj: Atspi.Accessible,
        inclusive: bool = False,
    ) -> bool:
        """Returns True if obj has a toolbar ancestor."""

        return AXUtilities._is_descendant(
            AXUtilities._CACHE.IS_TOOL_BAR_DESCENDANT,
            obj,
            AXUtilitiesRole.is_tool_bar,
            inclusive,
        )

    @staticmethod
    def is_tool_tip_descendant(
        obj: Atspi.Accessible,
        inclusive: bool = False,
    ) -> bool:
        """Returns True if obj has a tooltip ancestor."""

        return AXUtilities._is_descendant(
            AXUtilities._CACHE.IS_TOOL_TIP_DESCENDANT,
            obj,
            AXUtilitiesRole.is_tool_tip,
            inclusive,
        )

    @staticmethod
    def is_tree_or_tree_table_descendant(
        obj: Atspi.Accessible,
        inclusive: bool = False,
    ) -> bool:
        """Returns True if obj has a tree or tree table ancestor."""

        return AXUtilities._is_descendant(
            AXUtilities._CACHE.IS_TREE_OR_TREE_TABLE_DESCENDANT,
            obj,
            AXUtilitiesRole.is_tree_or_tree_table,
            inclusive,
        )

    @staticmethod
    def is_message_dialog(obj: Atspi.Accessible) -> bool:
        """Returns True if obj is a dialog that should be treated as a message dialog"""

        if not AXUtilitiesRole.is_dialog_or_alert(obj):
            return False

        if not AXObject.supports_collection(obj):
            widgets = AXUtilities.get_all_widgets(obj, exclude_push_button=True)
            return not widgets

        tokens = ["AXUtilities:", obj]
        result = True
        if AXUtilitiesCollection.has_scroll_pane(obj):
            tokens.append("is not a message dialog: has scroll pane")
            result = False
        elif AXUtilitiesCollection.has_split_pane(obj):
            tokens.append("is not a message dialog: has split pane")
            result = False
        elif AXUtilitiesCollection.has_tree_or_tree_table(obj):
            tokens.append("is not a message dialog: has tree or tree table")
            result = False
        elif AXUtilitiesCollection.has_combo_box_or_list_box(obj):
            tokens.append("is not a message dialog: has combo box or list box")
            result = False
        elif AXUtilitiesCollection.has_editable_object(obj):
            tokens.append("is not a message dialog: has editable object")
            result = False
        else:
            tokens.append("is believed to be a message dialog")

        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return result

    @staticmethod
    def is_redundant_object(obj1: Atspi.Accessible, obj2: Atspi.Accessible) -> bool:
        """Returns True if obj2 is redundant to obj1."""

        if obj1 == obj2:
            return False

        if AXObject.get_name(obj1) != AXObject.get_name(obj2) or AXObject.get_role(
            obj1,
        ) != AXObject.get_role(obj2):
            return False

        tokens = ["AXUtilities:", obj2, "is redundant to", obj1]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return True

    @staticmethod
    def _sort_by_child_index(object_list: list[Atspi.Accessible]) -> list[Atspi.Accessible]:
        """Returns the list of objects sorted according to child index."""

        def cmp(x, y):
            return AXObject.get_index_in_parent(x) - AXObject.get_index_in_parent(y)

        if not object_list or len(object_list) == 1:
            return object_list

        result = sorted(object_list, key=functools.cmp_to_key(cmp))

        first, second = result[0:2]
        if AXUtilitiesRole.is_radio_button(first) and AXObject.get_toolkit_name(first) == "gtk":
            # Gtk radio buttons are often in reverse order, except for when they're not.
            # See https://gitlab.gnome.org/GNOME/gtk/-/issues/7839.
            sorted_first, _sorted_second = AXUtilitiesComponent.sort_objects_by_position(
                [first, second]
            )
            if sorted_first != first:
                result.reverse()

        if object_list != result:
            tokens = ["AXUtilities: Original list", object_list]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            tokens = ["AXUtilities: Sorted list", result]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        return result

    @staticmethod
    def _get_set_members(
        obj: Atspi.Accessible,
        container: Atspi.Accessible,
    ) -> list[Atspi.Accessible]:
        """Returns the members of the container of obj"""

        if container is None:
            tokens = ["AXUtilities: Members of", obj, "not obtainable: container is None"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return []

        result = AXUtilitiesRelation.get_is_member_of(obj)
        if result:
            tokens = ["AXUtilities: Members of", obj, "in", container, "via member-of", result]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return AXUtilities._sort_by_child_index(result)

        result = AXUtilitiesRelation.get_is_node_parent_of(obj)
        if result:
            tokens = ["AXUtilities: Members of", obj, "in", container, "via node-parent-of", result]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return AXUtilities._sort_by_child_index(result)

        if AXUtilitiesRole.is_description_value(obj):
            previous_sibling = AXObject.get_previous_sibling(obj)
            while previous_sibling and AXUtilitiesRole.is_description_value(previous_sibling):
                result.append(previous_sibling)
                previous_sibling = AXObject.get_previous_sibling(previous_sibling)
            result.append(obj)
            next_sibling = AXObject.get_next_sibling(obj)
            while next_sibling and AXUtilitiesRole.is_description_value(next_sibling):
                result.append(next_sibling)
                next_sibling = AXObject.get_next_sibling(next_sibling)
            tokens = ["AXUtilities: Members of", obj, "in", container, "based on siblings", result]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return result

        if AXUtilitiesRole.is_menu_related(obj):
            result = list(AXObject.iter_children(container, AXUtilitiesRole.is_menu_related))
            tokens = ["AXUtilities: Members of", obj, "in", container, "based on menu role", result]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return result

        role = AXObject.get_role(obj)
        result = list(AXObject.iter_children(container, lambda x: AXObject.get_role(x) == role))
        tokens = ["AXUtilities: Members of", obj, "in", container, "based on role", result]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return result

    @staticmethod
    def get_set_members(obj: Atspi.Accessible) -> list[Atspi.Accessible]:
        """Returns the members of the container of obj."""

        container = AXObject.get_parent_checked(obj)
        result = AXUtilities._CACHE.get_members(container) or []

        if obj not in result:
            if result:
                tokens = ["AXUtilities:", obj, "not in cached members of", container, ":", result]
                debug.print_tokens(debug.LEVEL_INFO, tokens, True)

            result = AXUtilities._get_set_members(obj, container)
            AXUtilities._CACHE.set_members(container, result)

        # In a collapsed combobox, one can arrow to change the selection without showing
        # the items. In a listbox, items scrolled out of view lose the showing state but
        # are still valid members.
        def is_combo_box_or_list_box(x):
            return AXUtilitiesRole.is_combo_box(x) or AXUtilitiesRole.is_list_box(x)

        must_be_showing = not AXUtilitiesObject.find_ancestor(obj, is_combo_box_or_list_box)
        if not must_be_showing:
            return result

        filtered = list(filter(AXUtilitiesState.is_showing, result))
        if result != filtered:
            tokens = ["AXUtilities: Filtered non-showing:", set(result).difference(set(filtered))]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        return filtered

    @staticmethod
    def get_set_size(obj: Atspi.Accessible) -> int:  # pylint: disable=too-many-return-statements
        """Returns the total number of objects in this container."""

        result = AXObject.get_attribute(obj, "setsize", False)
        if isinstance(result, str) and result.isnumeric():
            return int(result)

        parent = AXObject.get_parent(obj)
        result = AXObject.get_attribute(parent, "setsize", False)
        if isinstance(result, str) and result.isnumeric():
            return int(result)

        if AXUtilitiesRole.is_table_row(obj):
            return AXTable.get_row_count(AXUtilitiesTable.get_table(obj))

        if AXUtilitiesRole.is_table_cell_or_header(obj) and not AXUtilitiesRole.is_table_row(
            parent,
        ):
            return AXTable.get_row_count(AXUtilitiesTable.get_table(obj))

        if AXUtilitiesRole.is_combo_box(obj):
            selected_children = AXUtilitiesSelection.get_selected_children(obj)
            if not selected_children:
                return -1
            if len(selected_children) == 1:
                obj = selected_children[0]

        if AXUtilitiesRole.is_list(obj) or AXUtilitiesRole.is_list_box(obj):
            obj = AXUtilities.get_list_item(obj)

        child_count = AXObject.get_child_count(parent)
        if child_count > 500:
            return child_count

        members = AXUtilities.get_set_members(obj)
        return len(members)

    @staticmethod
    def get_set_size_is_unknown(obj: Atspi.Accessible) -> bool:
        """Returns True if the total number of objects in this container is unknown."""

        if AXUtilitiesState.is_indeterminate(obj):
            return True

        attrs = AXObject.get_attributes_dict(obj, False)
        if attrs.get("setsize") == "-1":
            return True

        if AXObject.get_attribute(AXObject.get_parent(obj), "setsize", False) == "-1":
            return True

        if AXUtilitiesRole.is_table(obj):
            return attrs.get("rowcount") == "-1" or attrs.get("colcount") == "-1"

        return False

    @staticmethod
    def _get_table_row_position(obj: Atspi.Accessible) -> int:
        """Returns the position of a table row from ARIA attributes, or -1."""

        result = AXObject.get_attribute(obj, "rowindex", False)
        if not (isinstance(result, str) and result.isnumeric()) and AXObject.get_child_count(obj):
            cell = AXUtilities.get_table_cell_or_header(obj)
            result = AXObject.get_attribute(cell, "rowindex", False)

        if isinstance(result, str) and result.isnumeric():
            return int(result) - 1
        return -1

    @staticmethod
    def get_position_in_set(obj: Atspi.Accessible) -> int:
        """Returns the position of obj with respect to the number of items in its container."""

        result = AXObject.get_attribute(obj, "posinset", False)
        if isinstance(result, str) and result.isnumeric():
            return int(result) - 1

        if AXUtilitiesRole.is_table_row(obj):
            position = AXUtilities._get_table_row_position(obj)
            if position >= 0:
                return position

        if AXUtilitiesRole.is_table_cell_or_header(obj) and not AXUtilitiesRole.is_table_row(
            AXObject.get_parent(obj),
        ):
            return AXTable.get_cell_coordinates(obj)[0]

        if AXUtilitiesRole.is_combo_box(obj):
            selected_children = AXUtilitiesSelection.get_selected_children(obj)
            if not selected_children:
                return -1
            if len(selected_children) == 1:
                obj = selected_children[0]

        if AXObject.get_child_count(AXObject.get_parent(obj)) > 500:
            return AXObject.get_index_in_parent(obj)

        members = AXUtilities.get_set_members(obj)
        return members.index(obj) if obj in members else -1

    @staticmethod
    def has_explicit_name(obj: Atspi.Accessible) -> bool:
        """Returns True if obj has an author/app-provided name as opposed to a calculated name."""

        return AXObject.get_attribute(obj, "explicit-name") == "true"

    @staticmethod
    def has_visible_caption(obj: Atspi.Accessible) -> bool:
        """Returns True if obj has a visible caption."""

        if not (AXUtilitiesRole.is_figure(obj) or AXObject.supports_table(obj)):
            return False

        labels = AXUtilitiesRelation.get_is_labelled_by(obj)
        for label in labels:
            if AXUtilitiesRole.is_caption(label) and AXUtilitiesState.is_showing_and_visible(label):
                return True

        return False

    @staticmethod
    def has_non_inline_children(obj: Atspi.Accessible) -> bool:
        """Returns True if obj has children with block, flex, or table display."""

        for child in AXObject.iter_children(obj):
            display = AXObject.get_attribute(child, "display")
            if display in ("block", "flex", "table"):
                return True
        return False

    @staticmethod
    def get_displayed_label(obj: Atspi.Accessible) -> str:
        """Returns the displayed label of obj."""

        labels = AXUtilitiesRelation.get_is_labelled_by(obj)
        strings = [AXObject.get_name(label) or AXText.get_all_text(label) for label in labels]
        result = " ".join(strings)
        return result

    @staticmethod
    def get_displayed_description(obj: Atspi.Accessible) -> str:
        """Returns the displayed description of obj."""

        descriptions = AXUtilitiesRelation.get_is_described_by(obj)
        strings = [AXObject.get_name(desc) or AXText.get_all_text(desc) for desc in descriptions]
        result = " ".join(strings)
        return result

    @staticmethod
    def get_heading_level(obj: Atspi.Accessible) -> int:
        """Returns the heading level of obj."""

        if not AXUtilitiesRole.is_heading(obj):
            return 0

        use_cache = not AXUtilitiesState.is_editable(obj)
        attrs = AXObject.get_attributes_dict(obj, use_cache)

        try:
            value = int(attrs.get("level", "0"))
        except ValueError:
            tokens = ["AXUtilities: Exception getting value for", obj, "(", attrs, ")"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return 0

        return value

    @staticmethod
    def get_nesting_level(obj: Atspi.Accessible) -> int:
        """Returns the nesting level of obj."""

        def pred(x: Atspi.Accessible) -> bool:
            if AXUtilitiesRole.is_list_item(obj):
                return AXUtilitiesRole.is_list(AXObject.get_parent(x))
            return AXUtilitiesRole.have_same_role(obj, x)

        ancestors = []
        ancestor = AXUtilitiesObject.find_ancestor(obj, pred)
        while ancestor:
            ancestors.append(ancestor)
            ancestor = AXUtilitiesObject.find_ancestor(ancestor, pred)

        return len(ancestors)

    @staticmethod
    def _get_next_object(obj: Atspi.Accessible) -> Atspi.Accessible | None:
        """Returns the next object (depth first, unless there's a flows-to relation)"""

        if not AXObject.is_valid(obj):
            return None

        targets = AXUtilitiesRelation.get_flows_to(obj)
        for target in targets:
            if not AXObject.is_dead(target):
                return target

        index = AXObject.get_index_in_parent(obj) + 1
        parent = AXObject.get_parent(obj)
        while parent and not 0 < index < AXObject.get_child_count(parent):
            obj = parent
            index = AXObject.get_index_in_parent(obj) + 1
            parent = AXObject.get_parent(obj)

        if parent is None:
            return None

        next_object = AXObject.get_child(parent, index)
        if next_object == obj:
            tokens = ["AXUtilities:", obj, "claims to be its own next object"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return None

        return next_object

    @staticmethod
    def _get_previous_object(obj: Atspi.Accessible) -> Atspi.Accessible | None:
        """Returns the previous object (depth first, unless there's a flows-from relation)"""

        if not AXObject.is_valid(obj):
            return None

        targets = AXUtilitiesRelation.get_flows_from(obj)
        for target in targets:
            if not AXObject.is_dead(target):
                return target

        index = AXObject.get_index_in_parent(obj) - 1
        parent = AXObject.get_parent(obj)
        while parent and not 0 <= index < AXObject.get_child_count(parent) - 1:
            obj = parent
            index = AXObject.get_index_in_parent(obj) - 1
            parent = AXObject.get_parent(obj)

        if parent is None:
            return None

        previous_object = AXObject.get_child(parent, index)
        if previous_object == obj:
            tokens = ["AXUtilities:", obj, "claims to be its own previous object"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return None

        return previous_object

    @staticmethod
    def is_on_screen(obj: Atspi.Accessible, bounding_box: Atspi.Rect | None = None) -> bool:
        """Returns true if obj should be treated as being on screen."""

        AXObject.clear_cache(obj, False, "Updating to check if object is on screen.")

        tokens = ["AXUtilities: Checking if", obj, "is showing and visible...."]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if not AXUtilitiesState.is_showing_and_visible(obj):
            tokens = ["AXUtilities:", obj, "is not showing and visible. Treating as off screen."]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return False

        tokens = ["AXUtilities:", obj, "is showing and visible. Checking size and rect..."]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if AXUtilitiesComponent.has_no_size_or_invalid_rect(obj):
            tokens = ["AXUtilities: Rect of", obj, "is unhelpful. Treating as on screen."]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return True

        tokens = ["AXUtilities:", obj, "has size and a valid rect. Checking if off screen..."]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if AXUtilitiesComponent.object_is_off_screen(obj):
            tokens = ["AXUtilities:", obj, "is believed to be off screen."]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return False

        tokens = [
            "AXUtilities:",
            obj,
            "is not off screen. Checking",
            bounding_box,
            "intersection...",
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if bounding_box is not None and not AXUtilitiesComponent.object_intersects_rect(
            obj, bounding_box
        ):
            tokens = ["AXUtilities", obj, "not in", bounding_box, ". Treating as off screen."]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return False

        tokens = ["AXUtilities:", obj, "is believed to be on screen."]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return True

    @staticmethod
    def treat_as_leaf_node(obj: Atspi.Accessible) -> bool:
        """Returns True if obj should be treated as a leaf node."""

        if AXUtilitiesRole.children_are_presentational(obj):
            # In GTK, the contents of the page tab descends from the page tab.
            return not AXUtilitiesRole.is_page_tab(obj)

        role = AXObject.get_role(obj)
        if AXUtilitiesRole.is_combo_box(obj, role):
            return not AXUtilitiesState.is_expanded(obj)

        if AXUtilitiesRole.is_menu(obj, role) and not AXUtilitiesRole.has_role_from_aria(obj):
            return not AXUtilitiesState.is_expanded(obj)

        if AXObject.get_name(obj):
            return AXUtilitiesRole.is_link(obj, role) or AXUtilitiesRole.is_label(obj, role)

        return False

    @staticmethod
    def _get_on_screen_objects(
        root: Atspi.Accessible,
        bounding_box: Atspi.Rect | None = None,
    ) -> list:
        tokens = ["AXUtilities: Getting on-screen objects in", root, f"({hex(id(root))})"]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if not AXUtilities.is_on_screen(root, bounding_box):
            return []

        if AXUtilities.treat_as_leaf_node(root):
            return [root]

        if AXObject.supports_table(root) and AXObject.supports_selection(root):
            return list(AXUtilitiesTable.iter_visible_cells(root))

        objects = []
        root_name = AXObject.get_name(root)
        root_has_text = AXUtilitiesText.has_presentable_text(root)
        if root_name or AXObject.get_description(root) or root_has_text:
            objects.append(root)

        if bounding_box is None:
            bounding_box = AXComponent.get_rect(root)

        for i, child in enumerate(AXObject.iter_children(root)):
            tokens = [f"AXUtilities: Child {i} is", child, f"({hex(id(child))})"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)

            if (
                root_has_text
                and AXUtilitiesRole.is_static(child)
                and AXHypertext.get_character_offset_in_parent(child) == -1
            ):
                tokens = ["AXUtilities:", child, "text is inline in parent. Skipping as duplicate."]
                debug.print_tokens(debug.LEVEL_INFO, tokens, True)
                continue

            children = AXUtilities._get_on_screen_objects(child, bounding_box)
            objects.extend(children)
            if root_name and children and root in objects and root_name == AXObject.get_name(child):
                objects.remove(root)

        is_interactive = AXUtilitiesState.is_focusable(root) or AXUtilitiesAction.has_action(
            root, "click"
        )
        if not objects and is_interactive:
            objects = [root]

        return objects

    @staticmethod
    def get_on_screen_objects(
        root: Atspi.Accessible,
        bounding_box: Atspi.Rect | None = None,
    ) -> list:
        """Returns a list of onscreen objects in the given root."""

        result = AXUtilities._get_on_screen_objects(root, bounding_box)
        tokens = [f"AXUtilities: {len(result)} onscreen objects found in", root]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return result

    @staticmethod
    def active_descendant(obj: Atspi.Accessible) -> Atspi.Accessible | None:
        """Legacy table-cell code originally for managed descendants."""

        # TODO - JD: Determine what actually needs this support and why.

        if AXObject.is_dead(obj):
            return None

        if not AXUtilitiesRole.is_table_cell(obj):
            return obj

        if AXObject.get_name(obj):
            return obj

        # Its own text makes it the subject, not a managed descendant to defer to.
        if AXUtilitiesText.has_presentable_text(obj):
            return obj

        def pred(x: Atspi.Accessible) -> bool:
            return bool(AXObject.get_name(x) or AXText.get_all_text(x))

        child = AXUtilitiesObject.find_descendant(obj, pred)
        if child is not None:
            return child

        return obj

    @staticmethod
    def find_previous_object(
        obj: Atspi.Accessible,
        restrict_to: Atspi.Accessible | None = None,
    ) -> Atspi.Accessible | None:
        """Finds the object before this one, optionally restricted to a subtree."""

        result = AXUtilities._get_previous_object(obj)
        tokens = ["AXUtilities: Previous object for", obj, "is", result]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        if restrict_to is not None and not AXUtilitiesObject.is_ancestor(result, restrict_to, True):
            tokens = ["AXUtilities:", result, "is not a descendant of", restrict_to]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return None

        return result

    @staticmethod
    def find_next_object(
        obj: Atspi.Accessible,
        restrict_to: Atspi.Accessible | None = None,
    ) -> Atspi.Accessible | None:
        """Finds the object after this one, optionally restricted to a subtree."""

        result = AXUtilities._get_next_object(obj)
        tokens = ["AXUtilities: Next object for", obj, "is", result]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        if restrict_to is not None and not AXUtilitiesObject.is_ancestor(result, restrict_to, True):
            tokens = ["AXUtilities:", result, "is not a descendant of", restrict_to]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return None

        return result


for method_name, method in inspect.getmembers(AXUtilitiesAction, predicate=inspect.isfunction):
    setattr(AXUtilities, method_name, method)

for method_name, method in inspect.getmembers(AXUtilitiesApplication, predicate=inspect.isfunction):
    setattr(AXUtilities, method_name, method)

for method_name, method in inspect.getmembers(AXUtilitiesEvent, predicate=inspect.isfunction):
    setattr(AXUtilities, method_name, method)

for method_name, method in inspect.getmembers(AXUtilitiesHypertext, predicate=inspect.isfunction):
    setattr(AXUtilities, method_name, method)

for method_name, method in inspect.getmembers(AXUtilitiesRelation, predicate=inspect.isfunction):
    setattr(AXUtilities, method_name, method)

for method_name, method in inspect.getmembers(AXUtilitiesRole, predicate=inspect.isfunction):
    setattr(AXUtilities, method_name, method)

for method_name, method in inspect.getmembers(AXUtilitiesState, predicate=inspect.isfunction):
    setattr(AXUtilities, method_name, method)

for method_name, method in inspect.getmembers(AXUtilitiesSelection, predicate=inspect.isfunction):
    setattr(AXUtilities, method_name, method)

for method_name, method in inspect.getmembers(AXUtilitiesCollection, predicate=inspect.isfunction):
    if method_name.startswith("find"):
        setattr(AXUtilities, method_name, method)

for method_name, method in inspect.getmembers(AXUtilitiesMath, predicate=inspect.isfunction):
    setattr(AXUtilities, method_name, method)

for method_name, method in inspect.getmembers(AXUtilitiesObject, predicate=inspect.isfunction):
    setattr(AXUtilities, method_name, method)

for method_name, method in inspect.getmembers(AXUtilitiesComponent, predicate=inspect.isfunction):
    setattr(AXUtilities, method_name, method)

for method_name, method in inspect.getmembers(AXUtilitiesDocument, predicate=inspect.isfunction):
    setattr(AXUtilities, method_name, method)

for method_name, method in inspect.getmembers(AXUtilitiesTable, predicate=inspect.isfunction):
    setattr(AXUtilities, method_name, method)

for method_name, method in inspect.getmembers(AXUtilitiesText, predicate=inspect.isfunction):
    setattr(AXUtilities, method_name, method)



# Orca
#
# Copyright 2005-2009 Sun Microsystems Inc.
# Copyright 2010-2011 Orca Team
# Copyright 2011-2015 Igalia, S.L.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=too-many-return-statements
# pylint: disable=too-many-locals
# pylint: disable=too-many-branches
# pylint: disable=too-many-statements
# pylint: disable=too-many-boolean-expressions
# pylint: disable=too-many-arguments
# pylint: disable=too-many-positional-arguments

"""Produces speech presentation for accessible objects."""

from __future__ import annotations

from dataclasses import replace
from typing import TYPE_CHECKING, Any

import gi

gi.require_version("Atspi", "2.0")
from gi.repository import Atspi

from orca import (
    caret_navigator,
    debug,
    focus_manager,
    input_event_manager,
    messages,
    object_properties,
    speech_generator,
    speechserver,
)
from orca.ax_object import AXObject
from orca.ax_table import AXTable
from orca.ax_text import AXText
from orca.ax_utilities import AXUtilities
from orca.generator import ContentItem, ContentPosition, PresentationReason

if TYPE_CHECKING:
    from orca.speech_generator import SpeechGeneratorContext

    from . import script


class SpeechGenerator(speech_generator.SpeechGenerator):
    """Produces speech presentation for accessible objects."""

    # Type annotation to override the base class script type
    _script: script.Script

    @staticmethod
    def log_generator_output(func):
        """Decorator for logging."""

        def wrapper(*args, **kwargs):
            result = func(*args, **kwargs)
            tokens = [f"WEB SPEECH GENERATOR: {func.__name__}:", result]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return result

        return wrapper

    @log_generator_output
    def _generate_old_ancestors(self, obj: Atspi.Accessible) -> list[Any]:
        if self._get_content_position(obj).index > 0:
            return []

        return super()._generate_old_ancestors(obj)

    @log_generator_output
    def _generate_new_ancestors(self, obj: Atspi.Accessible) -> list[Any]:
        if (
            self._get_content_position(obj).index > 0
            and AXUtilities.find_ancestor(obj, AXUtilities.is_list) is None
        ):
            return []

        return super()._generate_new_ancestors(obj)

    def _generate_ancestors(
        self,
        obj: Atspi.Accessible,
        *,
        include_only: list | None = None,
        skip_roles: list | None = None,
        stop_at_roles: list | None = None,
        stop_after_roles: list | None = None,
    ) -> list[Any]:
        if not self._script.utilities.in_document_content(obj):
            return super()._generate_ancestors(
                obj,
                include_only=include_only,
                skip_roles=skip_roles,
                stop_at_roles=stop_at_roles,
                stop_after_roles=stop_after_roles,
            )

        result: list[Any] = []
        prior_obj = self._get_prior_obj()
        if prior_obj and AXObject.get_parent(prior_obj) == AXObject.get_parent(obj):
            return result

        if prior_obj and self._script.utilities.in_document_content(prior_obj):
            prior_doc = self._script.utilities.get_document_for_object(prior_obj)
            doc = self._script.utilities.get_document_for_object(obj)
            if prior_doc != doc and not self._script.utilities.get_document_for_object(doc):
                result = [super()._generate_accessible_name(doc)]

        if not AXUtilities.get_table(obj) and (
            AXUtilities.is_landmark(obj)
            or AXUtilities.is_math_related(obj)
            or AXUtilities.is_tool_tip(obj)
            or AXUtilities.is_status_bar(obj)
        ):
            return result

        if self._script.utilities.is_item_for_editable_combo_box(obj, prior_obj):
            return result

        stop_at_roles = [
            Atspi.Role.DOCUMENT_WEB,
            Atspi.Role.EMBEDDED,
            Atspi.Role.INTERNAL_FRAME,
            Atspi.Role.MATH,
            Atspi.Role.MENU_BAR,
        ]
        skip_roles = [
            Atspi.Role.PARAGRAPH,
            Atspi.Role.HEADING,
            Atspi.Role.LABEL,
            Atspi.Role.LINK,
            Atspi.Role.LIST_ITEM,
            Atspi.Role.TEXT,
        ]
        if AXUtilities.find_ancestor(obj, AXUtilities.is_editable_combo_box):
            skip_roles.append(Atspi.Role.COMBO_BOX)

        result.extend(
            super()._generate_ancestors(
                obj,
                include_only=include_only,
                stop_at_roles=stop_at_roles,
                skip_roles=skip_roles,
                stop_after_roles=[Atspi.Role.TOOL_BAR],
            ),
        )

        return result

    @log_generator_output
    def _generate_state_has_popup(self, obj: Atspi.Accessible) -> list[Any]:
        # TODO - JD: Can this be merged into the default's
        if self._only_speak_displayed_text():
            return []

        result: list[Any] = []
        attrs = AXObject.get_attributes_dict(obj)
        popup_type = attrs.get("haspopup", "false").lower()
        if popup_type == "dialog":
            result = [messages.HAS_POPUP_DIALOG]
        elif popup_type == "grid":
            result = [messages.HAS_POPUP_GRID]
        elif popup_type == "listbox":
            result = [messages.HAS_POPUP_LISTBOX]
        elif popup_type in ("menu", "true"):
            result = [messages.HAS_POPUP_MENU]
        elif popup_type == "tree":
            result = [messages.HAS_POPUP_TREE]

        if result:
            result.extend(self.voice(speech_generator.SYSTEM, obj=obj))
            return result

        return super()._generate_state_has_popup(obj)

    @log_generator_output
    def _generate_has_click_action(self, obj: Atspi.Accessible) -> list[Any]:
        if self._only_speak_displayed_text():
            return []

        if not self._script.utilities.in_document_content(obj):
            return []

        if AXUtilities.is_feed_article(obj):
            return []

        if not self._script.utilities.is_clickable_element(obj):
            return []

        result: list[Any] = [object_properties.STATE_CLICKABLE]
        result.extend(self.voice(speech_generator.SYSTEM, obj=obj))
        return result

    @log_generator_output
    def _generate_accessible_description(self, obj: Atspi.Accessible) -> list[Any]:
        if self._only_speak_displayed_text():
            return []

        if not self._script.utilities.in_document_content(obj):
            return super()._generate_accessible_description(obj)

        if not AXObject.is_valid(obj):
            return []

        # TODO - JD: Can this logic be moved into the default speech generator?
        if self._prefer_description_over_name(obj):
            return []

        if obj != self._context.focus:
            if AXUtilities.is_dialog_or_alert(obj, self._get_resolved_role()):
                return super()._generate_accessible_description(obj)
            if self._context.active_mode != focus_manager.MOUSE_REVIEW:
                return []

        if AXUtilities.is_text(obj, self._get_resolved_role()) and (
            self._get_reason() != PresentationReason.WHERE_AM_I_BASIC
        ):
            return []

        if (
            AXUtilities.is_link(obj, self._get_resolved_role())
            and caret_navigator.get_navigator().last_input_event_was_navigation_command()
        ):
            return []

        return super()._generate_accessible_description(obj)

    @log_generator_output
    def _generate_has_long_description(self, obj: Atspi.Accessible) -> list[Any]:
        if self._only_speak_displayed_text():
            return []

        if not self._script.utilities.in_document_content(obj):
            return []

        if not self._script.utilities.has_long_desc(obj):
            return []

        result: list[Any] = [object_properties.STATE_HAS_LONGDESC]
        result.extend(self.voice(speech_generator.SYSTEM, obj=obj))
        return result

    @log_generator_output
    def _generate_has_details(self, obj: Atspi.Accessible) -> list[Any]:
        if self._only_speak_displayed_text():
            return []

        if not self._script.utilities.in_document_content(obj):
            return super()._generate_has_details(obj)

        objs = AXUtilities.get_details(obj)
        if not objs:
            return []

        def obj_string(x):
            return str.strip(f"{AXObject.get_name(x)} {self.get_localized_role_name(x)}")

        to_present = ", ".join(set(map(obj_string, objs)))
        result: list[Any] = [object_properties.RELATION_HAS_DETAILS % to_present]
        result.extend(self.voice(speech_generator.SYSTEM, obj=obj))
        return result

    @log_generator_output
    def _generate_all_details(self, obj: Atspi.Accessible) -> list[Any]:
        if self._only_speak_displayed_text():
            return []

        objs: list[Atspi.Accessible] = []
        container = obj
        while container and not objs:
            objs = AXUtilities.get_details(container)
            container = AXObject.get_parent(container)

        if not objs:
            return []

        result: list[Any] = [object_properties.RELATION_HAS_DETAILS % ""]
        result.extend(self.voice(speech_generator.SYSTEM, obj=obj))

        result = []
        for o in objs:
            result.append(self.get_localized_role_name(o))
            result.extend(self.voice(speech_generator.SYSTEM, obj=obj))

            string = self._script.utilities.expand_eocs(o)
            if not string.strip():
                continue

            result.append(string)
            result.extend(self.voice(speech_generator.DEFAULT, obj=obj))
            result.extend(self._generate_pause(o))

        return result

    @log_generator_output
    def _generate_details_for(self, obj: Atspi.Accessible) -> list[Any]:
        if self._only_speak_displayed_text():
            return []

        if not self._script.utilities.in_document_content(obj):
            return super()._generate_details_for(obj)

        objs = AXUtilities.get_is_details_for(obj)
        if not objs:
            return []

        if self._is_leaving():
            return []

        if self._is_say_all():
            return []

        manager = input_event_manager.get_manager()
        if manager.last_event_was_forward_caret_navigation() and self._get_start_offset(obj):
            return []
        if (
            manager.last_event_was_backward_caret_navigation()
            and self._script.utilities.treat_as_text_object(obj)
            and self._get_end_offset(obj) not in [None, AXText.get_character_count(obj)]
        ):
            return []

        result: list[Any] = []
        for o in objs:
            string = (
                self._script.utilities.expand_eocs(o)
                or AXObject.get_name(o)
                or self.get_localized_role_name(o)
            )
            words = string.split()
            if len(words) > 5:
                words = [*words[0:5], "..."]

            result.append(object_properties.RELATION_DETAILS_FOR % " ".join(words))
            result.extend(self.voice(speech_generator.SYSTEM, obj=obj))
            result.extend(self._generate_pause(o))

        return result

    @log_generator_output
    def _generate_accessible_label_and_name(self, obj: Atspi.Accessible) -> list[Any]:
        if not self._script.utilities.in_document_content(obj):
            return super()._generate_accessible_label_and_name(obj)

        if (
            self._script.utilities.is_text_block_element(obj)
            and AXUtilities.has_presentable_text(obj)
            and not AXUtilities.is_landmark(obj, self._get_resolved_role())
            and not self._script.utilities.is_document(obj)
            and not AXUtilities.is_dpub(obj, self._get_resolved_role())
            and not AXUtilities.is_suggestion(obj, self._get_resolved_role())
        ):
            return []

        if obj == self._get_prior_obj() and AXUtilities.is_editable(obj):
            return []

        if AXUtilities.is_label(obj, self._get_resolved_role()) and AXObject.supports_text(obj):
            return []

        return super()._generate_accessible_label_and_name(obj)

    @log_generator_output
    def _generate_accessible_name(self, obj: Atspi.Accessible) -> list[Any]:
        if not self._script.utilities.in_document_content(obj):
            return super()._generate_accessible_name(obj)

        if (
            self._script.utilities.is_text_block_element(obj)
            and AXUtilities.has_presentable_text(obj)
            and not AXUtilities.is_landmark(obj, self._get_resolved_role())
            and not AXUtilities.is_dpub(obj, self._get_resolved_role())
            and not self._in_flat_review
        ):
            return []

        if AXUtilities.is_link(obj, self._get_resolved_role()) and self._get_content_string(obj):
            return []

        if AXUtilities.has_visible_caption(obj):
            return []

        if AXUtilities.is_figure(obj, self._get_resolved_role()) and self._get_ancestor_of():
            caption = self._get_ancestor_of()
            if not AXUtilities.is_caption(caption):
                caption = AXUtilities.find_ancestor(caption, AXUtilities.is_caption)
            if caption and obj in AXUtilities.get_is_label_for(caption):
                return []

        # TODO - JD: Can this logic be moved to the default speech generator?
        if AXObject.get_name(obj):
            if self._prefer_description_over_name(obj):
                result: list[Any] = [AXObject.get_description(obj)]
            else:
                name = AXObject.get_name(obj)
                if not AXUtilities.has_explicit_name(obj):
                    name = name.strip()
                result = [name]

            result.extend(self.voice(speech_generator.DEFAULT, obj=obj))
            return result

        return super()._generate_accessible_name(obj)

    @log_generator_output
    def _generate_accessible_label(self, obj: Atspi.Accessible) -> list[Any]:
        if not self._script.utilities.in_document_content(obj):
            return super()._generate_accessible_label(obj)

        label, _objects = self._script.utilities.infer_label_for(obj)
        if label:
            result: list[Any] = [label]
            result.extend(self.voice(speech_generator.DEFAULT, obj=obj))
            return result

        return []

    @log_generator_output
    def _generate_text_content(self, obj: Atspi.Accessible) -> list[Any]:
        targets = AXUtilities.get_is_label_for(obj)
        labelled_ancestor = (
            AXUtilities.find_ancestor(obj, lambda x: x in targets)
            if targets and AXUtilities.is_label(obj, self._get_resolved_role())
            else None
        )
        if labelled_ancestor is not None:
            prior_obj = self._get_prior_obj()
            entering = (
                prior_obj is None
                or AXUtilities.find_ancestor(prior_obj, lambda x: x == labelled_ancestor) is None
            )
            # Entering the group already announces it; only then is the legend's text redundant.
            if entering:
                return []
        return super()._generate_text_content(obj)

    @log_generator_output
    def _generate_leaving(self, obj: Atspi.Accessible) -> list[Any]:
        if self._only_speak_displayed_text():
            return []

        if not self._is_leaving():
            return []

        if self._script.utilities.in_document_content(
            obj,
        ) and not self._script.utilities.in_document_content(
            self._context.focus,
        ):
            result: list[Any] = [""]
            result.extend(self.voice(speech_generator.SYSTEM, obj=obj))
            return result

        return super()._generate_leaving(obj)

    @log_generator_output
    def _generate_new_radio_button_group(self, obj: Atspi.Accessible) -> list[Any]:
        # TODO - JD: The default speech generator"s method determines group membership
        # via the member-of relation. We cannot count on that here. Plus, radio buttons
        # on the web typically live in a group which is labelled. Thus the new-ancestor
        # presentation accomplishes the same thing. Unless this can be further sorted out,
        # try to filter out some of the noise....
        return []

    @log_generator_output
    def _generate_number_of_children(self, obj: Atspi.Accessible) -> list[Any]:
        if self._only_speak_displayed_text() or not self._context.verbose:
            return []

        # We handle things even for non-document content due to issues in
        # other toolkits (e.g. exposing list items to us that are not
        # exposed to sighted users)
        roles = [Atspi.Role.DESCRIPTION_LIST, Atspi.Role.LIST, Atspi.Role.LIST_BOX, "ROLE_FEED"]
        role = self._get_resolved_role(obj)
        if role not in roles:
            return super()._generate_number_of_children(obj)

        set_size = AXUtilities.get_set_size(AXObject.get_child(obj, 0))
        if set_size is None:
            if AXUtilities.is_description_list(obj, role):
                set_size = len(AXUtilities.get_description_list_terms(obj))
            elif AXUtilities.is_list_box(obj, role) or AXUtilities.is_list(obj, role):
                set_size = len(list(AXObject.iter_children(obj, AXUtilities.is_list_item)))

        if not set_size:
            return []

        if AXUtilities.is_description_list(obj):
            result: list[Any] = [messages.description_list_term_count(set_size)]
        elif role == "ROLE_FEED":
            result = [messages.feed_article_count(set_size)]
        else:
            result = [messages.list_item_count(set_size)]
        result.extend(self.voice(speech_generator.SYSTEM, obj=obj))
        return result

    def get_localized_role_name(
        self, obj: Atspi.Accessible, *, role: Atspi.Role | str | None = None
    ) -> str:
        if not self._script.utilities.in_document_content(obj):
            return super().get_localized_role_name(obj, role=role)

        role_description = AXObject.get_role_description(obj)
        if role_description:
            return role_description

        return super().get_localized_role_name(obj, role=role)

    @log_generator_output
    def _generate_real_active_descendant_displayed_text(
        self,
        obj: Atspi.Accessible,
    ) -> list[Any]:
        if not self._script.utilities.in_document_content(obj):
            return super()._generate_real_active_descendant_displayed_text(obj)

        rad = AXUtilities.active_descendant(obj)
        return self._generate_text_content(rad)

    @log_generator_output
    def _generate_accessible_role(self, obj: Atspi.Accessible) -> list[Any]:
        if not self._script.utilities.in_document_content(obj):
            return super()._generate_accessible_role(obj)

        # Do this check before the roledescription check, e.g. navigation within VSCode's editor.
        if obj == self._get_prior_obj() and not self._is_where_am_i():
            return []

        roledescription = AXObject.get_role_description(obj)
        if roledescription:
            result: list[Any] = [roledescription]
            result.extend(self.voice(speech_generator.SYSTEM, obj=obj))
            return result

        role = self._get_resolved_role(obj)
        start = self._get_start_offset(obj)
        end = self._get_end_offset(obj)

        ancestor_with_usable_role = self._get_ancestor_with_usable_role(obj)
        if not self._should_speak_role(obj):
            # Only the focused object absorbs an ancestor's role; an ancestor presented as
            # context must not borrow a role from further up the tree.
            if ancestor_with_usable_role and not self._is_ancestor():
                return self._generate_accessible_role(ancestor_with_usable_role)
            return []

        result = []
        mgr = input_event_manager.get_manager()
        is_editable = AXUtilities.is_editable(obj)
        if is_editable and not self._script.utilities.is_content_editable_with_embedded_objects(
            obj,
        ):
            if self._is_say_all() and start:
                return []
            if mgr.last_event_was_forward_caret_navigation() and start:
                return []
            if (
                mgr.last_event_was_backward_caret_navigation()
                and self._script.utilities.treat_as_text_object(obj)
                and end not in [None, AXText.get_character_count(obj)]
            ):
                return []
            result.append(self.get_localized_role_name(obj))
            result.extend(self.voice(speech_generator.SYSTEM, obj=obj))

        elif is_editable and self._script.utilities.is_document(obj):
            parent = AXObject.get_parent(obj)
            if (
                parent
                and not AXUtilities.is_editable(parent)
                and not mgr.last_event_was_caret_navigation()
            ):
                result.append(object_properties.ROLE_EDITABLE_CONTENT)
                result.extend(self.voice(speech_generator.SYSTEM, obj=obj))

        elif role == Atspi.Role.HEADING:
            if self._get_content_position(obj).last_for_object:
                level = AXUtilities.get_heading_level(obj)
                if level:
                    result.append(
                        object_properties.ROLE_HEADING_LEVEL_SPEECH
                        % {"role": self.get_localized_role_name(obj), "level": level},
                    )
                    result.extend(self.voice(speech_generator.SYSTEM, obj=obj))
                else:
                    result.append(self.get_localized_role_name(obj))
                    result.extend(self.voice(speech_generator.SYSTEM, obj=obj))

        elif self._script.utilities.is_link(obj):
            if AXUtilities.is_image(AXObject.get_parent(obj)):
                result.append(messages.IMAGE_MAP_LINK)
                result.extend(self.voice(speech_generator.SYSTEM, obj=obj))
            else:
                if self._script.utilities.has_useless_canvas_descendant(obj):
                    result.append(self.get_localized_role_name(obj, role=Atspi.Role.IMAGE))
                    result.extend(self.voice(speech_generator.SYSTEM, obj=obj))
                result.append(self.get_localized_role_name(obj))
                result.extend(self.voice(speech_generator.SYSTEM, obj=obj))

        else:
            result.append(self.get_localized_role_name(obj))
            result.extend(self.voice(speech_generator.SYSTEM, obj=obj))

        if ancestor_with_usable_role:
            result[1:1] = self._generate_accessible_role(ancestor_with_usable_role)

        return result

    @log_generator_output
    def _generate_position_in_list(self, obj: Atspi.Accessible) -> list[Any]:
        if AXUtilities.is_list_item(obj):
            if self._get_content_position(obj).index + 1 < self._get_content_position(obj).total:
                return []

        if not self._is_where_am_i():
            if self._get_prior_obj() == obj:
                return []

        return super()._generate_position_in_list(obj)

    @log_generator_output
    def _generate_state_unselected(self, obj: Atspi.Accessible) -> list[Any]:
        if not self._context.in_focus_mode:
            return []

        return super()._generate_state_unselected(obj)

    @log_generator_output
    def _generate_table_cell_contents(self, obj: Atspi.Accessible) -> list[Any]:
        """Generates the speech presentation for a single table cell."""

        result = super()._generate_table_cell_contents(obj)
        if not self._context.in_focus_mode:
            return result

        if self._context.announce_cell_coordinates:
            label = AXUtilities.get_label_for_cell_coordinates(obj)
            if label:
                result.append(label)
                result.extend(self.voice(speech_generator.SYSTEM, obj=obj))
                return result

            row, col = AXTable.get_cell_coordinates(obj)
            if AXUtilities.cell_row_changed(obj):
                result.append(messages.TABLE_ROW % (row + 1))
                result.extend(self.voice(speech_generator.SYSTEM, obj=obj))
            if AXUtilities.cell_column_changed(obj):
                result.append(messages.TABLE_COLUMN % (col + 1))
                result.extend(self.voice(speech_generator.SYSTEM, obj=obj))

        return result

    def generate_speech(
        self,
        obj: Atspi.Accessible,
        context: SpeechGeneratorContext,
        *,
        role: Atspi.Role | str | None = None,
        include_context: bool = True,
    ) -> list[Any]:
        if not self._script.utilities.in_document_content(obj):
            tokens = ["WEB:", obj, "is not in document content. Calling default speech generator."]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return super().generate_speech(obj, context, role=role, include_context=include_context)

        tokens = ["WEB: Generating speech for document object", obj]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        result: list[Any] = []
        if self._script.utilities.is_link(obj):
            role = Atspi.Role.LINK
        elif AXUtilities.is_custom_image(obj):
            role = Atspi.Role.IMAGE
        elif self._script.utilities.treat_as_div(obj, offset=self._get_start_offset(obj)):
            role = Atspi.Role.SECTION

        if context.prior_obj is None:
            document = self._script.utilities.get_top_level_document_for_object(obj)
            prior_context = self._script.utilities.get_prior_context(document)
            if prior_context:
                context = replace(context, prior_obj=prior_context[0])

        item = self._get_content_item(obj)
        start = item.start_offset if item is not None else 0
        end = item.end_offset if item is not None else -1
        language, dialect = self._script.utilities.get_language_and_dialect_for_substring(
            obj, start, end
        )
        context = replace(context, language=language, dialect=dialect)

        if not result:
            result = list(
                filter(
                    lambda x: x,
                    super().generate_speech(
                        obj,
                        context,
                        role=role,
                        include_context=include_context,
                    ),
                ),
            )

        tokens = ["WEB: Speech generation for document object", obj, "complete."]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return result

    def generate_contents(  # type: ignore[override]
        self,
        contents: list[tuple[Atspi.Accessible, int, int, str]],
        context: SpeechGeneratorContext,
    ) -> list[Any]:
        self._context = context
        return self._generate_web_contents(contents)

    def _generate_web_contents(
        self,
        contents: list[tuple[Atspi.Accessible, int, int, str]],
    ) -> list[Any]:
        if not contents:
            return []

        result: list[Any] = []
        contents = self._script.utilities.filter_contents_for_presentation(contents, True)
        tokens = ["WEB: Generating speech contents (length:", len(contents), ")"]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        announce_attrs = len(contents) > 1 and self._should_announce_attribute_changes(
            contents[0][0],
        )
        prev_attrs: dict[str, str] = {}
        system_voice = self.voice(speechserver.VoiceType.SYSTEM) if announce_attrs else []

        original_context = self._context
        base_position = original_context.content_position
        for i, content in enumerate(contents):
            obj, start, end, string = content
            tokens = [f"ITEM {i}: ", obj, f"start: {start}, end: {end} '{string}'"]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)

            # A content item can be just the line's trailing newline; skip it rather
            # than present it as "newline" or "blank".
            if string and not string.strip("\n"):
                continue

            if announce_attrs and i > 0:
                curr_attrs, run_start, run_end = AXText.get_text_attributes_at_offset(obj, start)
                exclude = AXUtilities.get_redundant_text_attributes(obj, run_start, run_end)
                result.extend(
                    [desc, *system_voice]
                    for desc in self._get_attribute_change_descriptions(
                        prev_attrs, curr_attrs, exclude
                    )
                )
                prev_attrs = curr_attrs
            elif announce_attrs:
                prev_attrs = AXText.get_text_attributes_at_offset(obj, start)[0]

            last_for_object = all(other[0] != obj for other in contents[i + 1 :])
            if base_position is not None:
                position = ContentPosition(
                    index=base_position.index + i,
                    total=base_position.total,
                    last_for_object=last_for_object,
                )
            else:
                position = ContentPosition(
                    index=i, total=len(contents), last_for_object=last_for_object
                )
            self._context = replace(
                self._context,
                content_item=ContentItem(start_offset=start, end_offset=end, string=string),
                content_position=position,
                content_subject=obj,
            )
            utterance = self.generate_speech(obj, self._context)
            if isinstance(utterance, list):

                def is_not_empty_list(x):
                    return not (isinstance(x, list) and not x)

                utterance = list(filter(is_not_empty_list, utterance))
            if utterance and utterance[0]:
                result.append(utterance)
                self._context = replace(self._context, prior_obj=obj)
        self._context = original_context

        if not result:
            if self._is_say_all() or not self._context.speak_blank_lines or self._is_ancestor():
                string = ""
            else:
                string = messages.BLANK
            # TODO - Why not [[string, *self.voice(speech_generator.DEFAULT)]] ?
            result = [string, *self.voice(speech_generator.DEFAULT)]

        return result

    def generate_line(
        self,
        obj: Atspi.Accessible,
        start_offset: int,
        end_offset: int,
        line: str,
        context: SpeechGeneratorContext,
    ) -> list[Any]:
        """Generates speech for a web document line via DOM-walking contents."""

        if not self._script.utilities.in_document_content(obj):
            return super().generate_line(obj, start_offset, end_offset, line, context)

        if AXUtilities.is_editable(obj) and "\ufffc" not in line:
            return super().generate_line(obj, start_offset, end_offset, line, context)

        document = self._script.utilities.get_top_level_document_for_object(obj)
        prior_context = self._script.utilities.get_prior_context(document=document)
        prior_obj = prior_context[0] if prior_context else None

        contents = self._script.utilities.get_line_contents_at_offset(
            obj,
            start_offset,
            use_cache=True,
        )
        return self.generate_contents(contents, replace(self._context, prior_obj=prior_obj))

    def generate_word(
        self,
        obj: Atspi.Accessible,
        offset: int,
        context: SpeechGeneratorContext,
    ) -> list[Any]:
        """Generates speech for a web document word via DOM-walking contents."""

        word_contents = self._script.utilities.get_word_contents_at_offset(
            obj,
            offset,
            use_cache=True,
        )
        if not word_contents:
            return []

        text_obj, start_offset, end_offset, word = word_contents[0]
        if "\ufffc" not in word and (
            AXUtilities.is_editable(text_obj) or AXUtilities.is_table_cell_or_header(text_obj)
        ):
            return self.generate_phrase(text_obj, start_offset, end_offset, word, context)

        prior_obj = text_obj if AXUtilities.is_text_input(text_obj) else None
        return self.generate_contents(
            word_contents,
            replace(self._context, prior_obj=prior_obj),
        )



# Copyright 2006, 2007, 2008, 2009 Brailcom, o.p.s.
#
# Author: Tomas Cerha <cerha@brailcom.org>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=broad-exception-caught
# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-locals

"""Provides an Orca speech server for Speech Dispatcher backend."""

from __future__ import annotations

import contextlib
import gc
import queue
import threading
import time
from typing import TYPE_CHECKING, Any

from gi.repository import GLib

from . import debug, guilabels, speechserver, systemd
from .acss import ACSS
from .speechserver import CapitalizationStyle, PunctuationStyle
from .ssml import SSML, SSMLCapabilities

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator
    from typing import ClassVar

    from . import input_event
    from .speechserver import VoiceFamily

try:
    import speechd
except Exception:
    _SPEECHD_AVAILABLE = False
else:
    _SPEECHD_AVAILABLE = True
    try:
        _CallbackType = speechd.CallbackType
    except AttributeError:
        _SPEECHD_VERSION_OK = False
    else:
        _SPEECHD_VERSION_OK = True


class SpeechServer(speechserver.SpeechServer):
    """Speech Dispatcher speech server for Orca."""

    _LOG_PREFIX: ClassVar[str] = "SPEECH DISPATCHER"

    @staticmethod
    def get_factory_name() -> str:
        return guilabels.SPEECH_DISPATCHER

    @staticmethod
    def get_speech_servers() -> list[speechserver.SpeechServer]:
        servers = []
        default = SpeechServer._get_speech_server(SpeechServer.DEFAULT_SERVER_ID)
        if default is not None:
            servers.append(default)
            for module in default.list_output_modules():
                server = SpeechServer._get_speech_server(module)
                if server is not None:
                    servers.append(server)
        return servers

    # *** Instance methods ***

    def __init__(self, server_id: str) -> None:
        super().__init__(server_id)
        self._client: Any = None
        self._probe_client: Any = None
        self._probe_pending: bool = False
        self._output_module: str | None = None
        self._current_synthesis_voice: str | None = None
        self._current_capitalization_style: str = CapitalizationStyle.NONE.value
        self._voice_families_cache: dict[
            tuple[str, str, str | None, int | None],
            list[tuple[str, str, str | None]],
        ] = {}
        if not _SPEECHD_AVAILABLE:
            msg = "ERROR: Speech Dispatcher is not available"
            debug.print_message(debug.LEVEL_WARNING, msg, True)
            return
        if not _SPEECHD_VERSION_OK:
            msg = "ERROR: Speech Dispatcher version 0.6.2 or later is required."
            debug.print_message(debug.LEVEL_WARNING, msg, True)
            return
        # The following constants must be initialized in runtime since they
        # depend on the speechd module being available.
        try:
            most = speechd.PunctuationMode.MOST
        except Exception:
            most = speechd.PunctuationMode.SOME
        self._punctuation_mode_map = {
            PunctuationStyle.ALL: speechd.PunctuationMode.ALL,
            PunctuationStyle.MOST: most,
            PunctuationStyle.SOME: speechd.PunctuationMode.SOME,
            PunctuationStyle.NONE: speechd.PunctuationMode.NONE,
        }
        self._callback_type_map = {
            speechd.CallbackType.BEGIN: speechserver.SayAllContext.PROGRESS,
            speechd.CallbackType.CANCEL: speechserver.SayAllContext.INTERRUPTED,
            speechd.CallbackType.END: speechserver.SayAllContext.COMPLETED,
            speechd.CallbackType.INDEX_MARK: speechserver.SayAllContext.PROGRESS,
        }

        self._default_voice_name = guilabels.SPEECH_DEFAULT_VOICE % server_id

        try:
            self._init()
        except Exception:
            debug.print_exception(debug.LEVEL_WARNING)
            msg = "ERROR: Speech Dispatcher service failed to connect"
            debug.print_message(debug.LEVEL_WARNING, msg, True)
        else:
            SpeechServer._active_servers[server_id] = self

    def _init(self) -> None:
        try:
            self._client = client = speechd.SSIPClient("Orca", component=self._id)
        except (speechd.SSIPCommunicationError, speechd.SpawnError) as error:
            msg = f"ERROR: Failed to connect to Speech Dispatcher: {error}"
            debug.print_message(debug.LEVEL_SEVERE, msg, True)
            self._client = None
            systemd.get_manager().set_status("Speech", f"not connected ({error})")
            return
        client.set_priority(speechd.Priority.MESSAGE)

        # The speechServerInfo setting is not connected to the speechServerFactory. As a result,
        # the user's chosen server (synthesizer) might be from spiel. Don't mutate self._id
        # here because it must stay consistent with the _active_servers key for proper cleanup.
        # _output_module (set via set_output_module) takes precedence because it reflects
        # what was explicitly configured; fall back to self._id for non-default servers.
        module = self._output_module
        if not module and self._id and self._id != self.DEFAULT_SERVER_ID:
            module = self._id
        if module:
            try:
                available_modules = client.list_output_modules()
                if module in available_modules:
                    client.set_output_module(module)
            except (AttributeError, speechd.SSIPCommandError):
                pass
        self._current_voice_properties = {}
        mode = self._punctuation_mode_map[self._current_punctuation_level]
        client.set_punctuation(mode)
        client.set_data_mode(speechd.DataMode.SSML)

    def update_capitalization_style(self, style: str) -> None:
        """Updates the capitalization style used by the speech server."""

        self._current_capitalization_style = style
        if self._client is None:
            return

        # "spell" is only applied per-character in speak_character() to
        # prevent the synthesizer from announcing capitals in words and phrases.
        if style == CapitalizationStyle.ICON.value:
            sd_style = "icon"
        else:
            sd_style = "none"

        try:
            self._client.set_cap_let_recogn(sd_style)
        except speechd.SSIPCommunicationError:
            msg = "SPEECH DISPATCHER: Connection lost. Trying to reconnect."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            self.reset()
            if self._client is not None:
                self._client.set_cap_let_recogn(sd_style)
        except Exception as error:
            msg = f"SPEECH DISPATCHER: set_cap_let_recogn failed: {error}"
            debug.print_message(debug.LEVEL_WARNING, msg, True)

    def update_punctuation_level(self, level: PunctuationStyle) -> None:
        """Punctuation level changed, inform this speechServer."""
        super().update_punctuation_level(level)
        if self._client is None:
            return
        mode = self._punctuation_mode_map[level]
        self._client.set_punctuation(mode)

    def _send_command(self, command: Callable[..., Any], *args: Any, **kwargs: Any) -> Any:
        try:
            return command(*args, **kwargs)
        except speechd.SSIPCommunicationError:
            msg = "SPEECH DISPATCHER: Connection lost. Trying to reconnect."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            self.reset()
            if self._client is None:
                return None
            try:
                return command(*args, **kwargs)
            except (AttributeError, speechd.SSIPCommandError) as error:
                msg = f"SPEECH DISPATCHER: Failed to reconnect: {error}"
                debug.print_message(debug.LEVEL_WARNING, msg, True)
                return None
        except Exception:
            return None

    def _set_rate(self, acss_rate: float) -> None:
        if self._client is None:
            return
        rate = int(2 * max(0, min(100, acss_rate)) - 100)
        self._send_command(self._client.set_rate, rate)

    def _set_pitch(self, acss_pitch: float) -> None:
        if self._client is None:
            return
        pitch = int(20 * max(0, min(10, acss_pitch)) - 100)
        self._send_command(self._client.set_pitch, pitch)

    def _set_pitch_range(self, acss_pitch_range: float) -> None:
        if self._client is None:
            return
        pitch_range = int(20 * max(0, min(10, acss_pitch_range)) - 100)
        self._send_command(self._client.set_pitch_range, pitch_range)

    def _set_volume(self, acss_volume: float) -> None:
        if self._client is None:
            return
        volume = int(20 * max(0, min(10, acss_volume)) - 100)
        self._send_command(self._client.set_volume, volume)

    def _set_family(self, acss_family: dict[str, Any] | None) -> None:
        if self._client is None:
            return

        lang, dialect = self._get_language_and_dialect(acss_family)
        if lang:
            self._send_command(self._client.set_language, lang)
            if dialect:
                # Try to set precise dialect
                self._send_command(self._client.set_language, lang + "-" + dialect)

        if acss_family is not None:
            if name := acss_family.get(speechserver.VoiceFamily.NAME):
                self._send_command(self._client.set_synthesis_voice, name)
                self._current_synthesis_voice = name

    def _debug_sd_values(self, prefix: str = "") -> None:
        if debug.debugLevel > debug.LEVEL_INFO:
            return

        try:
            if self._client is not None:
                sd_rate = self._send_command(self._client.get_rate)
                sd_pitch = self._send_command(self._client.get_pitch)
                sd_volume = self._send_command(self._client.get_volume)
                sd_language = self._send_command(self._client.get_language)
            else:
                sd_rate = sd_pitch = sd_volume = sd_language = "(client not available)"
        except Exception:
            sd_rate = sd_pitch = sd_volume = sd_language = "(exception occurred)"

        family = self._current_voice_properties.get(ACSS.FAMILY) or {}

        punctuation_style = self._current_punctuation_level.name

        msg = (
            f"SPEECH DISPATCHER: {prefix}\n"
            f"ORCA rate {self._current_voice_properties.get(ACSS.RATE)}, "
            f"pitch {self._current_voice_properties.get(ACSS.AVERAGE_PITCH)}, "
            f"pitch-range {self._current_voice_properties.get(ACSS.PITCH_RANGE)}, "
            f"volume {self._current_voice_properties.get(ACSS.GAIN)}, "
            f"language {self._get_language_and_dialect(family)[0]}, "
            f"punctuation: {punctuation_style}\n"
            f"SD rate {sd_rate}, pitch {sd_pitch}, volume {sd_volume}, language {sd_language}"
        )
        debug.print_message(debug.LEVEL_INFO, msg, True)

    def _apply_acss(self, acss: dict[str, Any] | None) -> None:
        super()._apply_acss(acss)
        self._apply_voice_to_backend()

    def _apply_voice_to_backend(self) -> None:
        """Push current voice properties to the speech-dispatcher connection."""

        if self._client is None:
            return
        current = self._current_voice_properties
        self._set_family(current.get(ACSS.FAMILY))
        self._set_rate(current.get(ACSS.RATE, 50))
        self._set_pitch(current.get(ACSS.AVERAGE_PITCH, 5.0))
        self._set_pitch_range(current.get(ACSS.PITCH_RANGE, 5.0))
        self._set_volume(current.get(ACSS.GAIN, 10.0))

    def _speak(self, text: str, acss: dict[str, Any] | None, **kwargs: Any) -> None:
        if isinstance(text, ACSS):
            text = ""

        ssml = SSML.markup_text(text, SSMLCapabilities.MARK)

        self._apply_acss(acss)
        self._debug_sd_values(f"Speaking '{ssml}' ")
        if self._client is not None:
            self._send_command(self._client.speak, ssml, **kwargs)

    def _say_all(
        self,
        iterator: Iterator[tuple[speechserver.SayAllContext, dict[str, Any]]],
        orca_callback: Callable[[speechserver.SayAllContext, int], None],
    ) -> bool:
        """Process another sayAll chunk.

        Called by the gidle thread.

        """
        try:
            context, acss = next(iterator)
        except StopIteration:
            pass
        else:

            def callback(callback_type, index_mark=None):
                # This callback is called in Speech Dispatcher listener thread.
                # No subsequent Speech Dispatcher interaction is allowed here,
                # so we pass the calls to the gidle thread.
                t = self._callback_type_map[callback_type]
                if t == speechserver.SayAllContext.PROGRESS:
                    if index_mark:
                        index = index_mark.split(":")
                        if len(index) >= 2:
                            start, end = index[0:2]
                            context.current_offset = context.start_offset + int(start)
                            context.current_end_offset = context.start_offset + int(end)
                            msg = (
                                f"SPEECH DISPATCHER: Got mark "
                                f"{context.current_offset}:{context.current_end_offset} / "
                                f"{context.start_offset}:{context.end_offset}"
                            )
                            debug.print_message(debug.LEVEL_INFO, msg, True)
                    else:
                        context.current_offset = context.start_offset
                        context.current_end_offset = None
                elif t == speechserver.SayAllContext.COMPLETED:
                    context.current_offset = context.end_offset
                    context.current_end_offset = None
                GLib.idle_add(orca_callback, context.copy(), t)
                if t == speechserver.SayAllContext.COMPLETED:
                    GLib.idle_add(self._say_all, iterator, orca_callback)

            self._speak(
                context.utterance,
                acss,
                callback=callback,
                event_types=list(self._callback_type_map.keys()),
            )
        return False  # to indicate, that we don't want to be called again.

    def _cancel(self) -> None:
        if self._client is not None:
            self._send_command(self._client.cancel)

    def get_voice_families(self) -> list[speechserver.VoiceFamily]:
        """Returns the list of voice families available in the current synthesizer."""

        voices: tuple[tuple[str, str, str | None], ...] = ()
        if self._client is not None:
            try:
                # This command is not available with older SD versions.
                list_synthesis_voices = self._client.list_synthesis_voices
            except AttributeError:
                pass
            else:
                try:
                    result = self._send_command(list_synthesis_voices)
                    if result:
                        voices = tuple(result)
                except Exception as error:
                    msg = f"SPEECH DISPATCHER: list_synthesis_voices failed: {error}"
                    debug.print_message(debug.LEVEL_WARNING, msg, True)

        return self._build_voice_families(voices)

    def speak(self, text: str | None = None, acss: dict[str, Any] | None = None) -> None:
        if not text:
            return

        if len(text) == 1:
            if (
                text.isupper()
                and text.isalpha()
                and self._current_capitalization_style == CapitalizationStyle.ICON.value
            ):
                # See https://gitlab.gnome.org/GNOME/orca/-/issues/657
                self._apply_acss(acss)
                self._speak(text, acss)
                return
            msg = f"SPEECH DISPATCHER: Speaking '{text}' as char"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            self._apply_acss(acss)
            if self._client is not None:
                self._send_command(self._client.char, text)
        else:
            msg = f"SPEECH DISPATCHER: Speaking '{text}' as string"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            self._speak(text, acss)

    def say_all(
        self,
        utterance_iterator: Iterator[tuple[speechserver.SayAllContext, dict[str, Any]]],
        progress_callback: Callable[[speechserver.SayAllContext, int], None],
    ) -> None:
        """Iterates through the given utterance_iterator, speaking each utterance."""

        GLib.idle_add(self._say_all, utterance_iterator, progress_callback)

    def speak_character(
        self,
        character: str,
        acss: dict[str, Any] | None = None,
        cap_style: CapitalizationStyle | None = None,
    ) -> None:
        """Speaks character."""

        self._apply_acss(acss)
        if cap_style == CapitalizationStyle.NONE:
            character = character.lower()

        name = character
        if not name or name == character:
            if cap_style == CapitalizationStyle.ICON:
                # char() does not reliably trigger the icon sound whereas
                # speak() does. See https://gitlab.gnome.org/GNOME/orca/-/issues/657
                self._speak(character, acss)
                return
            msg = f"SPEECH DISPATCHER: Speaking '{character}' as char"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            if self._client is not None:
                if cap_style == CapitalizationStyle.SPELL:
                    self._send_command(self._client.set_cap_let_recogn, "spell")
                self._send_command(self._client.char, character)
                if cap_style == CapitalizationStyle.SPELL:
                    self._send_command(self._client.set_cap_let_recogn, "none")
            return

        self.speak(name, acss)

    def speak_key_event(
        self,
        event: input_event.KeyboardEvent,
        acss: dict[str, Any] | None = None,
    ) -> None:
        """Speaks event."""

        event_string = event.get_key_name()
        locking_state_string = event.get_locking_state_string()
        event_string = f"{event_string} {locking_state_string}".strip()
        # Speaking this key event as a character makes the user experience more consistent.
        # Example: The user presses '.' and then presses BackSpace and expects the output to
        # be the same.
        if len(event_string) == 1:
            msg = f"SPEECH DISPATCHER: Speaking '{event_string}' as char"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            self._apply_acss(acss)
            if self._client is not None:
                self._send_command(self._client.char, event_string)
        else:
            msg = f"SPEECH DISPATCHER: Speaking '{event_string}' as string"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            self.speak(event_string, acss=acss)

    def clear_voice_families_cache(self) -> None:
        """Clear the cache for voice family lookups."""

        msg = "SPEECH DISPATCHER: Clearing voice families cache"
        debug.print_message(debug.LEVEL_INFO, msg, True)
        self._voice_families_cache.clear()

    def get_voice_families_for_language(
        self,
        language: str,
        dialect: str = "",
        variant: str | None = None,
        maximum: int | None = None,
    ) -> list[tuple[str, str, str | None]]:
        """Returns the families for language available in the current synthesizer."""

        if not language:
            language, dialect = self._get_language_and_dialect(None)
        elif not dialect:
            locale_language, locale_dialect = self._get_language_and_dialect(None)
            if language == locale_language:
                dialect = locale_dialect

        cache_key = (language, dialect, variant, maximum)
        if cache_key in self._voice_families_cache:
            msg = (
                f"SPEECH DISPATCHER: Returning cached result for language='{language}' "
                f"dialect='{dialect}' variant='{variant}'"
            )
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return self._voice_families_cache[cache_key]
        start = time.time()
        target_language, target_dialect = self._normalized_language_and_dialect(language, dialect)

        result = []
        if self._client is None:
            return result

        msg = (
            f"SPEECH DISPATCHER: Searching for language='{language}' "
            f"dialect='{dialect}' variant='{variant}'."
        )
        debug.print_message(debug.LEVEL_INFO, msg, True)

        language_with_dialect = f"{language}-{dialect}" if dialect else language
        try:
            voices = self._client.list_synthesis_voices(language_with_dialect, variant)
            msg = f"SPEECH DISPATCHER: Unfiltered voice list has {len(voices)} entries."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            if not voices:
                voices = self._client.list_synthesis_voices(language, variant)
                msg = f"SPEECH DISPATCHER: Unfiltered voice list (try 2) has {len(voices)} entries."
                debug.print_message(debug.LEVEL_INFO, msg, True)
        except (AttributeError, ValueError) as error:
            msg = f"SPEECH DISPATCHER: specifying language and variant failed: {error}"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            try:
                voices = self._client.list_synthesis_voices()
            except (AttributeError, ValueError, speechd.SSIPCommandError) as error2:
                msg = f"SPEECH DISPATCHER: Error listing synthesis voices: {error2}"
                debug.print_message(debug.LEVEL_WARNING, msg, True)
                return []
        except speechd.SSIPCommandError as error:
            msg = f"SPEECH DISPATCHER: Error listing synthesis voices: {error}"
            debug.print_message(debug.LEVEL_WARNING, msg, True)
            return []

        candidates, fallbacks = self._filter_voices_for_language(
            voices,
            target_language,
            target_dialect,
            maximum=maximum,
        )

        msg = (
            f"SPEECH DISPATCHER: Found {len(candidates)} match(es) for "
            f"language='{language}' dialect='{dialect}' variant='{variant}' "
            f"in {time.time() - start:.4f}s."
        )
        debug.print_message(debug.LEVEL_INFO, msg, True)
        if not candidates and fallbacks:
            msg = f"SPEECH DISPATCHER: No direct matches found, using {len(fallbacks)} fallback(s)."
            debug.print_message(debug.LEVEL_INFO, msg, True)
            candidates = fallbacks

        self._voice_families_cache[cache_key] = candidates
        return candidates

    def get_output_module(self) -> str:
        if self._client is None:
            return ""
        result = self._send_command(self._client.get_output_module)
        if result is not None:
            return result
        return ""

    def is_responsive(self, timeout: float = 2.0) -> bool:
        """Returns True if a probe round-trip to Speech Dispatcher returns within timeout."""

        # The probe runs on its own connection, not self._client: speechd's client puts no lock
        # around its send-command-then-read-reply exchange, so two threads sharing one connection
        # can read each other's replies. self._client stays main-thread-only.
        if self._client is None:
            return False

        # Called only on the main thread, and _probe_client is touched only by the probe thread, so
        # the check-and-set needs no lock. If a previous probe is still stuck on a wedged daemon,
        # report unresponsive rather than pile on another thread; the stuck probe clears the flag
        # when it unblocks.
        if self._probe_pending:
            return False

        self._probe_pending = True
        result: queue.Queue[bool] = queue.Queue()

        def probe() -> None:
            # Reuse the connection: reconnecting each probe is slow enough with some synthesizers
            # to blow the timeout even when the daemon is healthy.
            try:
                if self._probe_client is None:
                    self._probe_client = speechd.SSIPClient("Orca-probe", component=self._id)
                self._probe_client.get_output_module()
                result.put(True)
            except Exception:  # pylint: disable=broad-except
                if self._probe_client is not None:
                    with contextlib.suppress(Exception):
                        self._probe_client.close()
                    self._probe_client = None
                result.put(False)
            finally:
                self._probe_pending = False

        threading.Thread(target=probe, daemon=True).start()
        try:
            return result.get(timeout=timeout)
        except queue.Empty:
            msg = f"{self._LOG_PREFIX}: Speech server health check timed out"
            debug.print_message(debug.LEVEL_WARNING, msg, True)
            return False

    def set_output_module(self, module_id: str) -> None:
        """Set the speech output module to the specified provider."""

        # TODO - JD: This updates the output module, but not the the value of self._id.
        # That might be desired (e.g. self._id impacts what is shown in Orca preferences),
        # but it can be confusing.
        self._output_module = module_id
        if self._client is not None:
            self._send_command(self._client.set_output_module, module_id)
            self.clear_voice_families_cache()
            # Update the default voice name to match the new module
            self._default_voice_name = guilabels.SPEECH_DEFAULT_VOICE % module_id

    def stop(self) -> None:
        self._cancel()

    def shutdown(self) -> None:
        if self._probe_client is not None:
            with contextlib.suppress(Exception):
                self._probe_client.close()
            self._probe_client = None
        try:
            # Don't call _cancel() here because it can cut off messages we want to complete, such
            # as "screen reader off."
            if self._client is not None:
                self._client.close()
                # Set client to None to allow immediate garbage collection.
                self._client = None
                # Force garbage collection to clean up Speech Dispatcher objects immediately
                # This prevents hanging during Python's final garbage collection.
                gc.collect()
        except Exception as error:
            msg = f"SPEECH DISPATCHER: Error during shutdown of server {self._id}: {error}"
            debug.print_message(debug.LEVEL_WARNING, msg, True)
        finally:
            if self._id in SpeechServer._active_servers:
                del SpeechServer._active_servers[self._id]

    def reset(self) -> None:
        try:
            msg = f"SPEECH DISPATCHER: Resetting server {self._id}"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            if self._client is not None:
                self._client.close()
        except Exception as error:
            msg = f"SPEECH DISPATCHER: Error during reset of server {self._id}: {error}"
            debug.print_message(debug.LEVEL_WARNING, msg, True)
        finally:
            self.clear_voice_families_cache()
            self._init()
            if self._client is not None:
                systemd.get_manager().set_status("Speech", "reconnected")
            else:
                systemd.get_manager().set_status("Speech", "not connected")

    def list_output_modules(self) -> tuple[str, ...]:
        """Return names of available output modules as a tuple of strings."""

        if self._client is None:
            return ()
        try:
            return self._send_command(self._client.list_output_modules)
        except AttributeError:
            return ()
        except speechd.SSIPCommandError:
            return ()

    def get_voice_family(self) -> VoiceFamily:
        """Returns the current voice family as a VoiceFamily dictionary."""

        voice_name = self._current_synthesis_voice or self._default_voice_name
        language = ""
        try:
            language = self._send_command(self._client.get_language)
        except (AttributeError, speechd.SSIPCommandError) as error:
            msg = f"SPEECH DISPATCHER: Error getting language: {error}"
            debug.print_message(debug.LEVEL_WARNING, msg, True)

        lang_parts = language.partition("-")
        lang = lang_parts[0]
        dialect = lang_parts[2]
        return speechserver.VoiceFamily(
            {
                speechserver.VoiceFamily.NAME: voice_name,
                speechserver.VoiceFamily.LANG: lang,
                speechserver.VoiceFamily.DIALECT: dialect,
                speechserver.VoiceFamily.VARIANT: None,
            },
        )

    def set_voice_family(self, family: VoiceFamily) -> None:
        """Sets the voice family to family VoiceFamily dictionary."""

        if not family:
            return

        voice_name = family.get(speechserver.VoiceFamily.NAME, "")
        language = family.get(speechserver.VoiceFamily.LANG, "")
        dialect = family.get(speechserver.VoiceFamily.DIALECT, "")
        if dialect:
            language = f"{language}-{dialect}" if language else dialect

        # Try to set synthesis voice (individual voice like Nathan, Federica)
        if voice_name:
            try:
                self._send_command(self._client.set_synthesis_voice, voice_name)
                self._current_synthesis_voice = voice_name
            except AttributeError as error:
                msg = f"SPEECH DISPATCHER: Synthesis voice not supported: {error}"
                debug.print_message(debug.LEVEL_INFO, msg, True)
                self._current_synthesis_voice = None
            except speechd.SSIPCommandError as error:
                msg = f"SPEECH DISPATCHER: Error setting synthesis voice {voice_name}: {error}"
                debug.print_message(debug.LEVEL_WARNING, msg, True)

        if language:
            try:
                self._send_command(self._client.set_language, language)
            except speechd.SSIPCommandError as error:
                msg = f"SPEECH DISPATCHER: Error setting language {language}: {error}"
                debug.print_message(debug.LEVEL_WARNING, msg, True)
